#!/usr/bin/perl -w
#
###########################
# StorIQ Archiver V 0.9.0a
###########################
#
#        Copyright Kreode Technologies 2002
#        Copyright Intellique 2003, 2004
#
# This program is registered as IDDN.FR.001.180010.000.RP2002.000.10700
#
# May 3 2004
#
#
# vocabulary :
# "drive" refers to a storage device. A drive may have more than
# one actual tape/MO/disk drive/streamer (though not supported as now)
# "changer" refers to an automatic SCSI tape/drive exchanger/loader/jukebox
# of some sort.
#
# Messages :
# WARNING : the job will go on, but some things may go wrong (files in  silly place, DB not updated...)
# ERROR : the job will stop, but is partly done.
# FATAL : nothing will be done. The whole job was dropped.
# DIE : the program can't longer run. probably a misconfiguration, or a nasty bug.

use strict;
# Database access
use DBI;
# POSIX time
use POSIX qw(strftime);
# to send email reports
use Net::SMTP;
# to get an handle on ext programs STDERR
use IPC::Open3;
# checksum de fichiers
use Digest::MD5;

my $Version='0.9.0a';

# general variables
my ($VERBOSE,$DEBUG,$ONESHOT,$INITONLY,$EXITWHENDONE);

# general configuration variables:
my %ConfVar;
# drives infos from conf files
my @DriveVar;

# tape position;
my @DrivePos;
my $tapeposition=0;
# general use variables
my $key; my $val;
# backup devices counter
my $countdrives=-1;

# process list
my @processlist;

# Warning Messages to be inserted into MESSAGE field
my  $WarningMes;

# stop the program
my $term = 0;
my $child = 0;

$SIG{TERM} = sub { $term = 1 };

 sub REAPER
	{
	$child = 1 ;
	}

# get command line options
while (my $cliparm= shift)
	{
	if ($cliparm eq '-v')
		{$VERBOSE='TRUE';}
	if ($cliparm eq '-d')
		{$DEBUG='TRUE';}
	if ($cliparm eq '-i')
		{$INITONLY='TRUE';}
	if ($cliparm eq '-o')
		{$ONESHOT='TRUE';}
	if ($cliparm eq '-e')
		{$EXITWHENDONE='TRUE';}
	}

###################################
# start initialisation stage
###################################
#
# open logfile.
#this is defined here, because configuration files are opened later.
open ( STDERR , '>>../logs/StorIQ.log') or die "Can't open main log file! : $!";
open ( STDOUT , ">&STDERR") or die "Can't redirect STDOUT! : $!";
$|=1;

# logging
my $currtime = localtime (time ());
print STDOUT ("******************************\n$currtime StorIQ Archiver V $Version starting.\n");

# get config files

# logging
$currtime = localtime (time ());
print ("$currtime opening main config file.\n") if ($VERBOSE);

open ( MAINCONF , '<../etc/storiqarch.conf') or die "DIE : Can't open main config. file! : $!";

#declares variables from config file...
while (my $confline=<MAINCONF>)
	{
	if ($confline!~ /^#/ )
		{
		chomp $confline;
		if ($confline)
			{
			($key, $val)=split (/\s*=\s*/, $confline);
			$ConfVar{$key}=substr($val, 1, -1);
			print "main conf cl $key, val $val, $ConfVar{$key} \n" if ($DEBUG);
			}
		}

	}

close (MAINCONF);

#open drives conf. file describing installed hardware

# logging
$currtime = localtime (time ());
print ("$currtime opening drives config file.\n") if ($VERBOSE);

open ( DRIVECONF, '<../etc/drives.conf') or die "DIE : Can't open drives config. file! : $!";

while (my $confline=<DRIVECONF>)
	{
	if ($confline!~ /^#/ )
		{
		chomp $confline;
		if ($confline)
			{
			($key, $val)=split (/\s*=\s*/, $confline);
			$countdrives+=1 if ($key eq 'dr_name');
			$DriveVar[$countdrives]{$key}=substr($val, 1, -1);
			print "drive $countdrives cl $key, val $val, $DriveVar[$countdrives]{$key} \n" if ($DEBUG);
			}
		}

	}
close (DRIVECONF);

# for each drive, get info file and check parameters...
for (my $i=0; $i<=$countdrives; $i++)
     {
     # init possibly empty variables
     $DriveVar[$i]{'chgr_scdev'} = '' if (!$DriveVar[$i]{'chgr_scdev'});

     # logging
     $currtime = localtime (time ());
     my $thedrive='../etc/drivelist/'.$DriveVar[$i]{'dr_type'};
     print ("$currtime open drive $i type file $thedrive \n")  if ($VERBOSE);
     open ( DRIVEDESC, $thedrive) or warn ("WARNING : drive description file $thedrive not found!!\n");
     while (my $confline=<DRIVEDESC>)
	    {
	    if ($confline!~ /^#/ )
		{
		chomp $confline;
		if ($confline)
		    {
		    ($key, $val)=split (/\s*=\s*/, $confline);
		    $DriveVar[$i]{$key}=substr($val, 1, -1);
		    print "drive $i cl $key, val $val $DriveVar[$i]{$key} \n" if ($DEBUG);
		    }
		}
	    }
     close (DRIVEDESC);
     }

# save current directory
my $homedir=$ENV{'PWD'};
# chdir to root...
chdir('/');

# connect to database
# format: dbi:db type:db name:host:port
my $dbconn="dbi:$ConfVar{'DB_type'}:$ConfVar{'DB_name'}:$ConfVar{'DB_host'}:$ConfVar{'DB_port'}";
print "db connection $dbconn \n" if ($VERBOSE);
my $dbhdle = DBI->connect($dbconn, $ConfVar{'DB_user'}, $ConfVar{'DB_password'}) or die "Can't connect to the Database ! : $DBI::errstr\n";

# Check that tables exist : allfiles, alldrives, allslots, alltapes, allarchives, alljobs.
my @tables=$dbhdle->tables() or die "Can't get list of tables! : $DBI::errstr\n";
my $tablesexist=0;
foreach (@tables)
	  {
	  print "table $_ exists \n" if ($DEBUG);
	  $tablesexist+=1 if ($_ =~ m/allarchives|alltapes|allfiles|alljobs|alldrives|allslot|storiqusers/);
	  }
die ("Installation incomplete, missing tables in database!") if ($tablesexist!=7);

# cleanup old stuffs:
# set RUNNING jobs to ERROR
# mail the admin about errors.

# check hardware
$currtime = localtime (time ());
print ("$currtime getting hardware information.\n") if ($VERBOSE);

for (my $i=0; $i<=$countdrives; $i++)
	{
	my ($err, @TapeStatus, @ChgrStatus);
	# unknown position on tape....
	#~ $DrivePos[$i]=-1;
	if ($DriveVar[$i]{'dr_fam'} eq 'streamer')
		{
		($err, @TapeStatus)=&GetHwStatus($DriveVar[$i]{'dr_scdev'});
		print "drive $i error $err \ninfo :\n @TapeStatus \n" if ($DEBUG);
		if ($err)
			{
			$WarningMes.=$err;
			warn ("WARNING : error accessing drive, $err Won't update hardware information!\n");
			}
		}
	# changer status; right now can't manage anything but 1 drive!!!
	if ($DriveVar[$i]{'tp_changers'}>0)
		{
		($err, @ChgrStatus)=&GetHwStatus($DriveVar[$i]{'chgr_scdev'});
		print "drivechgr $i error $err tppos : $tapeposition\n changer :\n @ChgrStatus \n" if ($DEBUG);
		if ($err)
			{
			$WarningMes.=$err;
			warn ("WARNING : error accessing robot, $err Won't update hardware information!\n");
			}
		}

	my @TpInfos;
	foreach (@TapeStatus)
	       {
	       (my $trash, my $info)=split(/:/, $_);
	       $info =~ tr/a-zA-Z0-9\-_//cd ;
	       push (@TpInfos, $info);
	       }
	my @ChgrInfos;
	foreach (@ChgrStatus)
	       {
	       (my $trash, my $info)=split(/:/, $_);
	       $info =~ tr/a-zA-Z0-9\-_//cd ;
	       push (@ChgrInfos, $info);
	       }
	
	push (@TpInfos, ('','',''));
	push(@ChgrInfos, ('','',''));

	# refresh info in alldrives table
	# should be enhanced later to support more than one streamer per drive...
	if (!$err)
		{
		my $request="replace into alldrives ( DRIVEID, DRNAME, DRTYPE, TPFORMAT, DRDEV, DRSCDEV, CHGRDEV, TPDRVENDOR, TPDRMODEL, CHGRVENDOR, CHGRMODEL, DRSPEEDK, DRCYCLE, DRBLKSIZEK) values (";
		$request.="$i, '$DriveVar[$i]{'dr_name'}', '$DriveVar[$i]{'dr_type'}', '$DriveVar[$i]{'tp_type'}', '$DriveVar[$i]{'dr_dev'}', '$DriveVar[$i]{'dr_scdev'}', '$DriveVar[$i]{'chgr_scdev'}', '$TpInfos[1]', '$TpInfos[2] $TpInfos[0]', '$ChgrInfos[1]', '$ChgrInfos[2] $ChgrInfos[0]', '$DriveVar[$i]{'tp_speed'}', '$DriveVar[$i]{'tp_cycle'}', '$DriveVar[$i]{'tp_blk'}')";

		print "Request : $request \n\n" if ($DEBUG);
		# then insert appropriate informations in the drivelist Database.
		my $sth=$dbhdle->prepare($request);
		$sth->execute or (die "Unable to execute query ! : $dbhdle->errstr \n");
		$sth->finish;
		}
	}

print ("$currtime database disconnect.\n") if ($VERBOSE);
my $rc  = $dbhdle->disconnect;

###################################
# end of initialisation stage
###################################


###############
# get tape drives and changers status : tape position, etc.
$currtime = localtime (time ());
print ("$currtime getting tapes and loaders informations.\n") if ($VERBOSE);
FORK :
for (my $i=0; $i<=$countdrives; $i++)
	{
	# $pid 
	my $pid=fork();
	if (!$pid)
		{
		MAINJOBLOOP :
		while (!$term)
			{
			print ("[$$] each process connect to database.\n") if ($VERBOSE);
			# reconnect for each job..
			print "new db connection $dbconn \n" if ($VERBOSE);
			$dbhdle = DBI->connect($dbconn, $ConfVar{'DB_user'}, $ConfVar{'DB_password'}) or die "[$$]Can't connect to the Database ! : $DBI::errstr\n";
			my ($err, @TapeInfo, @ChgrInfo);
			# arrays to store the datas
			my %Slots = (SLOTNUM=>'',
						DRIVE=>'',
						TAPEIN=>'',
						TAPEID=>'',
						TAPELABEL=>'',
						STATUS=>'',
						FROMSLOT=>'');        # all slots parms
			
			
			my @Loadedtapes;  # which tapes are currently available
			my %Streamers;        # what slot are streamers and what tape is in...
			my $tpid;        # tape ID from tape itself
			my $tplabel;        #TAPE LABEL (read from tape or DB).
			my $tpspace;         # space available on tape in drive in Megabytes
			my @TheFiles;        # files to save, verify or restore
			my @workedFiles;        # list of files actually saved/restored/listed
			my $newtape;        # in case the current tape is new, set it to the new tape number

			$term=1 if ($ONESHOT);
			$term=1  if ($INITONLY);

			# delete drive info (but for single tape drives).
			# empty slots table (that means : confide in your mtx program...);
			if (!$err && $DriveVar[$i]{'tp_changers'}>0)
				{
				my $request="DELETE from allslots where DRIVEID=$i";
				my $sth=$dbhdle->do($request);
				}

			# here set the global tape position to the current drive position (fork is done)
			#~ $tapeposition=$DrivePos[$i];

			# get info
			@TapeInfo=();
			if ($DriveVar[$i]{'dr_fam'} eq 'streamer')
				{
				($err, @TapeInfo)=&GetTpInfo($DriveVar[$i]{'dr_dev'});
				print "[$$] Tape drive $i error : $err\n \ninfo :\n @TapeInfo \n" if ($DEBUG);
				if ($err)
					{
					$WarningMes.=$err;
					warn ("[$$] WARNING: error accessing drive, $err Won't update hardware information!\n");
					}
				}
			# changer status; right now can't manage anything but 1 changer!!!
			if ($DriveVar[$i]{'tp_changers'}>0)
				{
				($err, @ChgrInfo)=&GetChgrStatus($DriveVar[$i]{'chgr_scdev'});
				print "[$$] Changer $i info :\n @ChgrInfo error : $err\n" if ($DEBUG);
				if ($err)
					{
					$WarningMes.=$err;
					warn ("[$$] WARNING: error accessing robot, $err Won't update hardware information!\n");
					}
				}
			else
				{
				# it's a simple drive (no changer)
				$Streamers{'SLOT'}=0;
				# generate a "ChangerInfo"

				if ($DriveVar[$i]{'dr_fam'} eq 'fixeddisk')
					{
					# it's a harddrive...
					print "[$$] drive $i is a fixed disk.\n" if ($VERBOSE);
					$Slots{'TAPEIN'}='Y';
					$Slots{'TAPELABEL'}=$DriveVar[$i]{'dr_name'};
					# generate and use a virtual cassette of name   "name of drive".
					# does the virtual tape exist already?
					my $request="select TAPEID, TPSPACE from alltapes WHERE TPNAME='$DriveVar[$i]{'dr_name'}'";
					my $sth=$dbhdle->prepare($request);
					$sth->execute;
					print "[$$] request : $request\n" if ($DEBUG);
					if (my ($row1, $tpsize)=$sth->fetchrow_array)
						{
						# it's not a new tape...
						$Slots{'TAPEID'}=$row1;
						$Slots{'STATUS'}='UNUSED';
						$tpspace=$tpsize;
						print "[$$] TAPEID : $row1 TPSIZE : $tpsize \n" if ($DEBUG);
						# what's the status of the loaded tape?
						#it is...
						if (($tpsize>0) and ($tpsize<$DriveVar[$i]{'tp_maxsize'}))
							{
							$Slots{'STATUS'}='INUSE';
							$Slots{'STATUS'}='FULL' if ($tpsize<($DriveVar[$i]{'tp_maxsize'}*.05));
							}
						elsif ($tpsize<0)
							{$Slots{'STATUS'}='ERROR';}
						}
					else
						{
						# it's a new tape
						$Slots{'STATUS'}='UNUSED';
						$tpspace=$DriveVar[$i]{'tp_maxsize'};
						# get a new TAPEID (not secured... needs locking)
						my $request="select TAPEID from alltapes ORDER BY TAPEID DESC LIMIT 1";
						my $sth=$dbhdle->prepare($request);
						$sth->execute;
						($newtape)=$sth->fetchrow_array();
						$newtape++;
						$Slots{'TAPEID'}=$newtape;
						$request="INSERT INTO alltapes ( TAPEID, TPNAME, TPLABEL, TPSPACE, TPFORMAT) values (";
						$request.=" '$newtape', '$Slots{'TAPELABEL'}', '$Slots{'TAPELABEL'}', '$DriveVar[$i]{'tp_maxsize'}', '$DriveVar[$i]{'tp_type'}' )";
						print "[$$] Request : $request \n\n" if ($DEBUG);
						$sth=$dbhdle->do($request) or warn ( $WarningMes.="[$$] WARNING: Can't create tape status for tape $Slots{'TAPEID'}, labeled $Slots{'TAPELABEL'}, request $request!\n");
						}
					push(@Loadedtapes, $Slots{'TAPEID'}) ;
					}
				elsif ($DriveVar[$i]{'dr_fam'} eq 'streamer')
					{
					# it's a tape drive
					print "[$$] drive $i is a simple tape drive.\n" if ($VERBOSE);
					# is there a tape inside?
					if (@TapeInfo)
						{
						$Slots{'SLOTNUM'}=0;
						$Slots{'TAPEIN'}='N';
						if (join("\n", @TapeInfo)=~ m/ONLINE/)
							{
							# there's a tape.
							my $tpid;
							$Slots{'TAPEIN'}='Y';
							# if the tape is already identified, don't read the header!
							my $request="select TAPEID from allslots WHERE( DRIVEID='$i' AND SLOTNUM='0' )";
							my $sth=$dbhdle->prepare($request);
							$sth->execute;
							($tpid)=$sth->fetchrow_array;
							print "[$$] request : $request\ntape label from DB $tpid\n" if ($DEBUG);
							# identify the tape from header
							($err, $tpid)=&GetTapeID($DriveVar[$i]{'dr_dev'}) if (!$tpid);
							print "[$$] tape label from tape $tpid \n, err $err, tppos : $tapeposition\n" if ($DEBUG);
							warn "[$$] Getting tape ID, error : $err \n" if ($err);

							# then what tape is it?
							$request="select TAPEID, TPLABEL, TPSPACE from alltapes WHERE TAPEID='$tpid'";
							$sth=$dbhdle->prepare($request);
							$sth->execute;
							print "[$$] request : $request\n" if ($DEBUG);
							my ($row1, $row2, $tpsize)=$sth->fetchrow_array;
							if ($row1>0)
								{
								$Slots{'TAPELABEL'}=$row2;
								# it's not a new tape...
								$Slots{'TAPEID'}=$row1;
								$Slots{'STATUS'}='UNUSED';
								$tpspace=$tpsize;
								print "[$$] TAPEID : $row1 TPSIZE : $tpsize \n" if ($DEBUG);
								# what's the status of the loaded tape?
								#it is...
								if (($tpsize>0) and ($tpsize<$DriveVar[$i]{'tp_maxsize'}))
									{
									$Slots{'STATUS'}='INUSE';
									$Slots{'STATUS'}='FULL' if ($tpsize<($DriveVar[$i]{'tp_maxsize'}*.05));
									}
								elsif ($tpsize<0)
									{$Slots{'STATUS'}='ERROR';}
								}
							else
								{
								# it's a new tape
								$Slots{'STATUS'}='UNUSED';
								$tpspace=$DriveVar[$i]{'tp_maxsize'};
								# get a new TAPEID (not secured... needs locking)
								my $request="select TAPEID from alltapes ORDER BY TAPEID DESC LIMIT 1";
								my $sth=$dbhdle->prepare($request);
								$sth->execute;
								($newtape)=$sth->fetchrow_array();
								$Slots{'TAPEID'}=$newtape+1;
								# create TAPE HEADER and TAPE END.
								($err, $newtape)=&NewTape($DriveVar[$i]{'dr_dev'}, $Slots{TAPEID});
								warn "[$$] could not create tape $Slots{TAPEID} header!\n" if ($err);
								($err, $newtape)=&CreateEndMark($DriveVar[$i]{'dr_dev'}, $Slots{TAPEID});
								warn "[$$] could not create tape $Slots{TAPEID} END header!\n" if ($err);
								$request="INSERT INTO alltapes ( TAPEID, TPNAME, TPLABEL, TPSPACE, TPFORMAT) values ( '$Slots{'TAPEID'}', '$Slots{'TAPELABEL'}', '$Slots{'TAPELABEL'}', '$DriveVar[$i]{'tp_maxsize'}', '$DriveVar[$i]{'tp_type'}' )" if (!$err);
								print "[$$] Request : $request \n\n" if ($DEBUG);
								$sth=$dbhdle->do($request) or warn ( $WarningMes.="[$$] WARNING: Can't create tape status for tape $Slots{'TAPEID'}, labeled $Slots{'TAPELABEL'}, request $request!\n");
								}
							}

						}
					else
						{
						# no info from mt...
						warn "no information about drive!\n";
						}
					push(@Loadedtapes, $Slots{'TAPEID'}) ;
					}
				else
					{
					# it's a single CD/WORM drive
					print "[$$] drive $i is a removable/WORM drive.\n" if ($VERBOSE);
					}

				# then  create the slot...
				my $request="replace into allslots (DRIVEID, SLOTNUM, SLOTTYPE, TAPEIN, TAPEID, TPLABEL, STATUS, FROMSLOT) values (";
				$request.="$i, '$Slots{'SLOTNUM'}', 'DRIVE', '$Slots{'TAPEIN'}', '$Slots{'TAPEID'}', '$Slots{'TAPELABEL'}', '$Slots{'STATUS'}', '$Slots{'FROMSLOT'}')";
				print "[$$] Request : $request \n\n" if ($DEBUG);

				# store drive info in database : table for all slots
				my $sth=$dbhdle->do($request) or (die "Unable to execute query ! : $dbhdle->errstr \n");
				}

			shift(@ChgrInfo);
			# here review all slots from changer
			foreach (@ChgrInfo)
				{
				(my $p1, my $p2, my $p3)=split(/:/, $_, 3);
				my $tpsize;
				$Slots{'TAPEID'}=0;
				$Slots{'TAPELABEL'}='';
				if ($p3)
					{
					$p3=~ s/VolumeTag\s*=\s*// ;
					$p3=~ s/\s*//g;
					chomp $p3;
					$Slots{'TAPELABEL'}=$p3;
					# either it's a  cleaning tape (according to main config file)
					$Slots{'TAPEID'}=-1 if ($Slots{'TAPELABEL'}=~ m/$ConfVar{'cleaning_tape'}/);
					}
				$Slots{'SLOTNUM'}=$p1;
				$Slots{'SLOTNUM'}=~ tr/0-9//cd;
				$Slots{'FROMSLOT'}='';
				$Slots{'STATUS'}='UNKNOWN';
				if ($p2=~ m/Full/)
					{
					$Slots{'TAPEIN'}='Y';
					# what's the TAPE ID?  0=> unknown tape (unreadable label, no label, etc), -1=>CLEANING tape
					my $request="select TAPEID, TPSPACE from alltapes WHERE TPLABEL='$Slots{'TAPELABEL'}'";
					my $sth=$dbhdle->prepare($request);
					$sth->execute;
					print "[$$] request : $request\n" if ($DEBUG);
					if (my ($row1, $row2)=$sth->fetchrow_array)
						{
						$Slots{'TAPEID'}=$row1;
						$Slots{'STATUS'}='UNUSED';
						print "[$$] TAPEID : $row1 TPSIZE : $row2 tppos : $tapeposition\n" if ($DEBUG);
						$tpsize=$row2;
						# what's the status of the loaded tape?
						#it is...
						if (($row2>0) and ($row2<$DriveVar[$i]{'tp_maxsize'}))
							{
							$Slots{'STATUS'}='INUSE';
							$Slots{'STATUS'}='FULL' if ($row2<($DriveVar[$i]{'tp_maxsize'}*.05));
							}
						elsif ($row2<0)
							{$Slots{'STATUS'}='ERROR';}
						}
					else
						{
						# it's a new tape
						warn "[$$] WARNING : No readable label on tape\n" if (!$Slots{'TAPELABEL'});
						$Slots{'STATUS'}='UNUSED';
						# get a new TAPEID (not secured... needs locking)
						my $request="select TAPEID from alltapes ORDER BY TAPEID DESC LIMIT 1";
						my $sth=$dbhdle->prepare($request);
						$sth->execute;
						($newtape)=$sth->fetchrow_array();
						$newtape++;
						$Slots{'TAPEID'}=$newtape;
						$request="INSERT INTO alltapes ( TAPEID, TPLABEL, TPSPACE, TPFORMAT) values (";
						$request.=" '$newtape', '$Slots{'TAPELABEL'}', '$DriveVar[$i]{'tp_maxsize'}', '$DriveVar[$i]{'tp_type'}' )";
						print "[$$] Request : $request \n\n" if ($DEBUG);
						$sth=$dbhdle->do($request) or warn ( $WarningMes.="[$$] WARNING : Can't create tape status for tape $Slots{'TAPEID'}, labeled $Slots{'TAPELABEL'}, request $request!\n");
						}
					}
				else
					{
					$Slots{'TAPEIN'}='N';
					}
				push(@Loadedtapes, $Slots{'TAPEID'}) ;

				if ($p1=~ m/Data Transfer/)
					{
					$Slots{'SLOTTYPE'}='DRIVE';
					$Slots{'FROMSLOT'}=$p2;
					$Slots{'FROMSLOT'}=~ tr/0-9//cd;
					$Streamers{'SLOT'}=$Slots{'SLOTNUM'};
					$Streamers{'FROMSLOT'}=$Slots{'FROMSLOT'};
					$Streamers{'TAPEIN'}='Y' if ($p2=~ m/Full/);
					$tpspace=$tpsize;
					}
				elsif ($p1=~ m/Storage/)
					{
					$Slots{'SLOTTYPE'}='SLOT';
					}
				else
					{
					$Slots{'SLOTTYPE'}='INOUT';
					}

				my $request="insert into allslots (DRIVEID, SLOTNUM, SLOTTYPE, TAPEIN, TAPEID, TPLABEL, STATUS, FROMSLOT) values (";
				$request.="$i, '$Slots{'SLOTNUM'}', '$Slots{'SLOTTYPE'}', '$Slots{'TAPEIN'}', '$Slots{'TAPEID'}', '$Slots{'TAPELABEL'}', '$Slots{'STATUS'}', '$Slots{'FROMSLOT'}')";
				print "[$$] Request : $request \n\n" if ($DEBUG);

				# store drive info in database : table for all slots
				my $sth=$dbhdle->do($request) or (die "Unable to execute query ! : $dbhdle->errstr \n");
				# $sth->finish;
				}

			##############################
			# check if there's a programmed job
			# first select missed jobs
			my @fields=qw( PJOBID REPEATMODE REPEATCOUNT MESSAGE REPORTTO);
			print "[$$] FIELDS @fields\n" if ($DEBUG);
			my %rows;
			my $sth;
			$currtime = localtime (time ());
			print "[$$] looking for MISSED programmed jobs at $currtime\n" if ($VERBOSE);
			my $request="SELECT ".join(', ',@fields)." FROM progjobs WHERE ( DRIVEID='$i' AND JENDTIME<NOW() AND STATUS='TODO' )";
			print "[$$] Request : $request \n\n" if ($DEBUG);

			if (!$INITONLY)
				{
				$sth=$dbhdle->prepare($request);
				$sth->execute;
				@rows{@fields}=();
				$sth->bind_columns(map { \$rows{$_} } @fields);
				while ($sth->fetch())
					{
					print "\n[$$] \@rows : ".join(":",%rows)."\n\n" if ($DEBUG && %rows );
					if ($rows{PJOBID})
						{
						# update missed jobs
						$request="UPDATE progjobs SET STATUS='MISSED', MESSAGE='$rows{MESSAGE}MISSED repetition number $rows{REPEATCOUNT} at $currtime\n'";
						if (($rows{REPEATMODE} ne 'None') and ($rows{REPEATCOUNT}!=0))
							{
							my $repeat=$rows{REPEATCOUNT};
							$repeat-- if ($repeat>0);
							my $period="1 DAY";
							if ($rows{REPEATMODE} eq 'Week')
								{$period="7 DAY";}
							elsif ($rows{REPEATMODE} eq 'Month')
								{$period="1 MONTH";}

							$request.=", REPEATCOUNT=$repeat, JSTTIME=DATE_ADD(JSTTIME, INTERVAL $period), JENDTIME=DATE_ADD(JSTTIME, INTERVAL $period)";
							}
						$request.=" WHERE PJOBID=$rows{PJOBID}";
						my $sth2=$dbhdle->do($request) or warn ($WarningMes.="[$$] ERROR : can't update prog. job $rows{PJOBID} to MISSED!\n request failed : $request\n");
						# generates alerts for missed jobs
						if (($rows{REPORTTO}) and ($rows{REPORTTO}!~ m/ADMIN/ ))
							{
							$rows{REPORTTO}.="\nADMIN\n";
							}
						elsif (!$rows{REPORTTO})
							{
							$rows{REPORTTO}.="ADMIN\n";
							}

						# split all emails
						my @themails=split(/\n/, $rows{REPORTTO});
						foreach (@themails)
							{
							my $themail=$_;
							$themail=$ConfVar{'adm_mail'} if ($themail eq 'ADMIN');
							#should check emails are valid someday.
							# SndEmail (destmail, destname, subject, body);
							&SndEmail($themail,$_,"StorIQ Archiver $rows{JOBTYPE} job $rows{PJOBID} MISSED","$rows{JOBTYPE} programmed Job $rows{PJOBID}, with description :\n\"$rows{DESCR}\"\n was marked MISSED at $currtime.\n");
							}
						}
					}
				}

			#select prog jobs
			@fields=qw( PJOBID JOBTYPE FROMDIR CMDOPT WHAT TAPEID DRIVEID REPORTTO ISARCHIVE CREATOR ARCHID ARCHNAME DESCR  REPEATMODE REPEATCOUNT GENJOB);
			print "[$$] FIELDS @fields\n" if ($DEBUG);
			$currtime = localtime (time ());
			print "[$$] looking for TODO programmed jobs at $currtime\n" if ($VERBOSE);
			$request="SELECT ".join(', ',@fields)." FROM progjobs WHERE ( JSTTIME < NOW() AND DRIVEID='$i' AND ( STATUS='TODO' or STATUS='MISSED') )";
			print "[$$] Request : $request \n\n" if ($DEBUG);


			if (!$INITONLY)
				{
				$sth=$dbhdle->prepare($request);
				$sth->execute;
				@rows{@fields}=();
				$sth->bind_columns(map { \$rows{$_} } @fields);
				$sth->execute;
				while ($sth->fetch())
					{
					print "[$$] programmed jobs:n\@rows : ".join(":",%rows)."\n\n" if ($DEBUG && %rows );

					if ($rows{PJOBID})
						{
						# for each  starting programmed job, insert one or two new jobs with max priority
						if ($rows{JOBTYPE} eq 'INCSAVE')
							{
							my @insfields=qw(JOBTYPE PRIORITY CMDOPT TAPEID DRIVEID WHAT REPORTTO STATUS MESSAGE ISARCHIVE CREATOR ARCHID ARCHNAME DESCR );
							my $insjob;
							# COMPARE if first, else VERIFY
							my $jobmode='COMPARE';
							$jobmode='VERIFY' if ($rows{GENJOB});

							$request="INSERT INTO alljobs ( ".join(', ',@insfields).") VALUES (";
							$request.="'$jobmode', '-20', '$rows{CMDOPT}', '$rows{TAPEID}', '$rows{DRIVEID}', '', '$rows{REPORTTO}', 'TODO', 'Generated from PJob $rows{PJOBID} at $currtime\n', '$rows{ISARCHIVE}', '$rows{CREATOR}', '$rows{ARCHID}', '$rows{ARCHNAME}', '$rows{DESCR}')";
							print "[$$] Request : $request \n\n" if ($DEBUG);
							if (my $sth2=$dbhdle->do($request))
								{
								# update GENJOB...
								$request="SELECT JOBID FROM alljobs WHERE MESSAGE='Generated from PJob $rows{PJOBID} at $currtime\n'";
								print "[$$] Request : $request \n\n" if ($DEBUG);
								my $sth3=$dbhdle->prepare($request);
								$sth3->execute;
								($insjob)=$sth->fetchrow_array or warn "[$$] WARNING : can't get new job!\n request : $request\n";
								$rows{GENJOB}.="$insjob\n";
								}
							# then DIFFSAVE
							$request="INSERT INTO alljobs ( ".join(', ',@insfields).") VALUES (";
							$request.="'DIFFSAVE', '-20', '$rows{CMDOPT}', '$rows{TAPEID}', '$rows{DRIVEID}', 'insjob', '$rows{REPORTTO}', 'TODO', 'Generated from PJob $rows{PJOBID} at $currtime\n', 'N', '$rows{CREATOR}', '$rows{ARCHID}', '$rows{ARCHNAME}', '$rows{DESCR}')";
							print "[$$] Request : $request \n\n" if ($DEBUG);
							if (my $sth2=$dbhdle->do($request))
								{
								# update GENJOB...
								$request="SELECT JOBID FROM alljobs WHERE MESSAGE='Generated from PJob $rows{PJOBID} at $currtime\n'";
								print "[$$] Request : $request \n\n" if ($DEBUG);
								my $sth3=$dbhdle->prepare($request);
								$sth3->execute;
								($insjob)=$sth3->fetchrow_array or warn "[$$] WARNING : can't get new job!\n request : $request\n";
								$rows{GENJOB}.="$insjob\n";
								}
							}
						else
							{
							my @insfields=qw(JOBTYPE PRIORITY FROMDIR CMDOPT WHAT TAPEID DRIVEID REPORTTO STATUS MESSAGE ISARCHIVE CREATOR ARCHID ARCHNAME DESCR );
							$request="INSERT INTO alljobs ( ".join(', ',@insfields).") VALUES (";
							$request.="'$rows{JOBTYPE}', '-20', '$rows{FROMDIR}', '$rows{CMDOPT}', '$rows{WHAT}', '$rows{TAPEID}', '$rows{DRIVEID}', '$rows{REPORTTO}', 'TODO', 'Generated from PJob $rows{PJOBID} at $currtime\n', '$rows{ISARCHIVE}', '$rows{CREATOR}', '$rows{ARCHID}', '$rows{ARCHNAME}', '$rows{DESCR}')";
							print "[$$] Request : $request \n\n" if ($DEBUG);
							if (my $sth2=$dbhdle->do($request))
								{
								# update GENJOB...
								$request="SELECT JOBID FROM alljobs WHERE MESSAGE='Generated from PJob $rows{PJOBID} at $currtime\n'";
								print "[$$] Request : $request \n\n" if ($DEBUG);
								my $sth3=$dbhdle->prepare($request);
								$sth3->execute;
								my ($insjob)=$sth3->fetchrow_array or warn "[$$] WARNING : can't get new job!\n request : $request\n";
								$rows{GENJOB}.="$insjob\n";
								}
							else
								{
								warn ($WarningMes.="[$$] ERROR : can't create $rows{JOBTYPE} job from prog. job $rows{PJOBID}!\n request failed : $request\n");
								}

							}
						# for each started progjob, update GENJOB, REPEATCOUNT if REPEATCOUNT>0
						# update DATETIME +1 day if REPEAT Day, +1 Week if REPEAT Week, +1 Month if REPEAT Month.
						$request="UPDATE progjobs SET GENJOB='$rows{GENJOB}'";
						if (($rows{REPEATMODE} ne 'None') and ($rows{REPEATCOUNT}!=0))
							{
							my $repeat=$rows{REPEATCOUNT};
							$repeat-- if ($repeat>0);
							my $period="1 DAY";
							if ($rows{REPEATMODE} eq 'Week')
								{$period="7 DAY";}
							elsif ($rows{REPEATMODE} eq 'Month')
								{$period="1 MONTH";}

							$request.=", REPEATCOUNT=$repeat, JSTTIME=DATE_ADD(JSTTIME, INTERVAL $period), JENDTIME=DATE_ADD(JSTTIME, INTERVAL $period)";
							print "[$$] Request : $request \n\n" if ($DEBUG);
								
							# set all progjobs with REPEACOUNT=0 to DONE and others to TODO (to revert MISSED to TODO)
							if ($repeat==0)
								{
								$request.=", STATUS='DONE'";
								}
							else
								{
								$request.=", STATUS='TODO'";
								}
							}
						else
							{
							$request.=", STATUS='DONE'";
							}
						$request.=" WHERE PJOBID=$rows{PJOBID}";
						print "[$$] Request : $request \n\n" if ($DEBUG);
							

						my $sth2=$dbhdle->do($request) or warn "[$$] WARNING : can't update progjob for job $rows{PJOBID}!\n request : $request\n";

						}
					}
				}


			##############################
			# get a job
			# insert a lock to avoid race on database. Tough job!
			#grab the first available (TODO) job in the DB for the current drive!
			# $rows 0: JOBID 1: JOBTYPE 2: FROMDIR 3: CMDOPT 4: WHAT 5: TAPEID 6: DRIVEID 7: REPORTTO 8: STATUS
			# 9: MESSAGE 10 : ISARCHIVE 11: CREATOR 12: ARCHID 13: ARCHID 14: DESCR 15: JSTTIME 16: JENDTIME
			@fields=qw( JOBID JOBTYPE FROMDIR CMDOPT WHAT TAPEID DRIVEID REPORTTO STATUS MESSAGE ISARCHIVE CREATOR ARCHID ARCHNAME DESCR JSTTIME JENDTIME );
			print "[$$] FIELDS @fields\n" if ($DEBUG);

			$request="SELECT ".join(', ',@fields)." FROM alljobs WHERE STATUS='TODO' AND DRIVEID='$i' ORDER BY PRIORITY, JOBID LIMIT 1";
			print "[$$] Request : $request \n\n" if ($DEBUG);

			if (!$INITONLY)
				{
				$sth=$dbhdle->prepare($request);
				$sth->execute;
				@rows{@fields}=();
				$sth->bind_columns(map { \$rows{$_} } @fields);
				if ($sth->fetch())
					{
					print "\n[$$] \@rows : ".join(":",%rows)."\n\n" if ($DEBUG && %rows );
					}
				$sth->finish;
				}

			################################
			# if there's a job...
			DOJOB :
			if ($rows{JOBID})
				{
				$currtime = localtime (time ());
				print ("$currtime start job $rows{JOBID}...\n");

				# set the job status to RUNNING
				my $SQLtime=strftime "%Y-%m-%d %H:%M:%S", localtime;
				$request="UPDATE alljobs set STATUS='RUNNING', JSTTIME='$SQLtime' where JOBID=$rows{JOBID}";
				$sth=$dbhdle->do($request) or &ExitOnError ( $WarningMes.="[$$] FATAL : can't update DB job status ! \n request : $request", $rows{JOBID});

				# ready for tape dance?
				## is the necessary tape available? (in the changers)

				my $tapeishere=-1;
				for (my $i=0; $i< @Loadedtapes; $i++)
					{
					if ($rows{TAPEID}==$Loadedtapes[$i])
						{
						$tapeishere=$i;
						last;
						}
					}

				print "[$$] tape $rows{TAPEID} is in slot $tapeishere \n" if ($VERBOSE);

				if ($tapeishere==-1)
					{
					# then step down the job and run another one.
					my $SQLtime=strftime "%Y-%m-%d %H:%M:%S", localtime;
					$request="UPDATE alljobs set STATUS='DELAYED', MESSAGE='JOB DELAYED : TAPE $rows{TAPEID} not available!\n$WarningMes', JENDTIME='$SQLtime' where JOBID=$rows{JOBID}";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->do($request) or &ExitOnError ( $WarningMes.="[$$] FATAL : can't update DB job status ! \n request : $request", $rows{JOBID});
					print "Tape $rows{TAPEID} not available!\n Skipping job.\n";
					last;
					}
				### Is the tape loaded?
				if ($tapeishere!=$Streamers{'SLOT'})
					{
					print "[$$] tape not in streamer $Streamers{'SLOT'} \n" if ($DEBUG);
					### tape not loaded :
					### unload the loaded tape to its previous slot
					if ($Streamers{'TAPEIN'} eq 'Y')
						{
						if (!$Streamers{'FROMSLOT'})
							{
							# select the first available slot
							$request="SELECT SLOTNUM FROM allslots WHERE TAPEIN='N' LIMIT 1";
							print "[$$] Request : $request \n\n" if ($DEBUG);
							$sth=$dbhdle->prepare($request);
							$sth->execute;
							($Streamers{'FROMSLOT'})=$sth->fetchrow_array or warn "[$$] WARNING : no empty slot!\n";
							$sth->finish;
							}

						my ($err, $res)=&UnloadTape($DriveVar[$i]{'chgr_scdev'},  $DriveVar[$i]{'dr_dev'}, $Streamers{'SLOT'},  $Streamers{'FROMSLOT'}) if ($DriveVar[$i]{'tp_changers'}>0);
						#update changer info
						if (!$err)
							{
							$request="UPDATE allslots set TAPEIN='N', TAPEID='0', STATUS='UNKNOWN' where (DRIVEID='$i' and SLOTNUM='$Streamers{'SLOT'}')";
							print "[$$] Request : $request \n\n" if ($DEBUG);
							my $sth=$dbhdle->do($request) or warn ( $WarningMes.="[$$] WARNING : Can't update changer status for tape $rows{TAPEID}!\n $request failed.\n");
							$request="UPDATE allslots set TAPEIN='Y', TAPEID='$Loadedtapes[$Streamers{'SLOT'}]', STATUS='INUSE' where (DRIVEID='$i' and SLOTNUM='$Streamers{'FROMSLOT'}')";
							print "[$$] Request : $request \n\n" if ($DEBUG);
							$sth=$dbhdle->do($request) or warn ( $WarningMes.="[$$] WARNING : Can't update changer status for tape $rows{TAPEID}!\n $request failed.\n");
							}
						else
							{
							&ExitOnError($WarningMes.="[$$] FATAL : can't unload tape from drive $i to slot $Streamers{'FROMSLOT'}! Got : $err\n", $rows{JOBID});
							}
						}

					### now load the tape
					my ($err, $res)=&LoadTape($DriveVar[$i]{'chgr_scdev'}, $DriveVar[$i]{'dr_dev'}, $Streamers{'SLOT'}, $tapeishere) if ($DriveVar[$i]{'tp_changers'}>0);
					if (!$err)
						{
						$Streamers{'FROMSLOT'}=$tapeishere;
						$request="UPDATE allslots set TAPEIN='N', TAPEID='0', TPLABEL='', STATUS='UNKNOWN' where (DRIVEID='$i' and SLOTNUM='$tapeishere')";
						print "[$$] Request : $request \n\n" if ($DEBUG);
						my $sth=$dbhdle->do($request) or warn ( $WarningMes.="[$$] WARNING : Can't update changer status for tape $rows{TAPEID}!\n $request failed.\n");
						$request="SELECT TPLABEL, TPSPACE FROM alltapes WHERE TAPEID='$Loadedtapes[$tapeishere]'";
						print "[$$] Request : $request \n\n" if ($DEBUG);
						$sth=$dbhdle->prepare($request);
						$sth->execute;
						($tplabel, $tpspace)=$sth->fetchrow_array or warn "[$$] WARNING : no such tape!\n";
						$sth->finish;
						$request="UPDATE allslots set TAPEIN='Y', TAPEID='$Loadedtapes[$tapeishere]', TPLABEL='$res', STATUS='INUSE' where (DRIVEID='$i' and SLOTNUM='$Streamers{'SLOT'}')";
						print "[$$] Request : $request \n\n" if ($DEBUG);
						$sth=$dbhdle->do($request) or warn ( $WarningMes.="[$$] WARNING : Can't update changer status for tape $rows{TAPEID}!\n $request failed.\n");
						}
					else
						{
						&ExitOnError("[$$] FATAL : can't load tape into drive $i! Got : $err\n", $rows{JOBID});
						}

					# check its label when loading.
					($err, $tpid)=&GetTapeID($DriveVar[$i]{'dr_dev'}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
					print "[$$] tape label from tape $tpid, looking for $rows{TAPEID} \n, err $err, tppos : $tapeposition\n" if ($DEBUG);
					$WarningMes.="[$$] Getting tape ID, error :".$err if ($err);
					if ($tpid<=0)
						{
						# if the tape is UNUSED, label it...
						($err,$newtape)=&NewTape($DriveVar[$i]{'dr_dev'}, $rows{TAPEID}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						&ExitOnError($WarningMes.="[$$] FATAL : can't create new tape $rows{TAPEID} : $err\n", $rows{JOBID})  if ($err);
						$tpid=$rows{TAPEID};
						print "[$$] NEWTAPE, $newtape, tppos : $tapeposition\n" if ($DEBUG);
						}

					# in case of a wrong label, unload the tape!
					elsif ($tpid!=$rows{TAPEID})
						{
						# where to unload?
						if (!$Streamers{'FROMSLOT'})
							{
							# select the first available slot
							$request="SELECT SLOTNUM FROM allslots WHERE TAPEIN='N' LIMIT 1";
							print "[$$] Request : $request \n\n" if ($DEBUG);
							$sth=$dbhdle->prepare($request);
							$sth->execute;
							($Streamers{'FROMSLOT'})=$sth->fetchrow_array or warn "[$$] WARNING : no empty slot!\n";
							$sth->finish;
							}

						# Unload this tape
						my ($err, $res)=&UnloadTape($DriveVar[$i]{'chgr_scdev'},  $DriveVar[$i]{'dr_dev'}, $Streamers{'SLOT'},  $Streamers{'FROMSLOT'}) if ($DriveVar[$i]{'tp_changers'}>0);

						#update changer info
						if (!$err)
							{
							$request="UPDATE allslots set TAPEIN='N', TAPEID='0', STATUS='UNKNOWN' where (DRIVEID='$i' and SLOTNUM='$Streamers{'SLOT'}')";
							print "[$$] Request : $request \n\n" if ($DEBUG);
							my $sth=$dbhdle->do($request) or warn ( $WarningMes.="[$$] WARNING : Can't update changer status for tape $rows{TAPEID}!\n $request failed.\n");
							$request="UPDATE allslots set TAPEIN='Y', TAPEID='$Loadedtapes[$Streamers{'SLOT'}]', STATUS='INUSE' where (DRIVEID='$i' and SLOTNUM='$Streamers{'FROMSLOT'}')";
							print "[$$] Request : $request \n\n" if ($DEBUG);
							$sth=$dbhdle->do($request) or warn ( $WarningMes.="[$$] WARNING : Can't update changer status for tape $rows{TAPEID}!\n $request failed.\n");
							}
						else
							{
							&ExitOnError($WarningMes.="[$$] FATAL : can't unload wrong tape from drive $i to slot $Streamers{'FROMSLOT'}! Got : $err\n", $rows{JOBID});
							}

						#Exit
						&ExitOnError ( $WarningMes.="[$$] FATAL : wrong mark on tape : tape $rows{TAPEID}, label $tplabel, mark $tpid !!!", $rows{JOBID});
						}
					}
				else
					{
					$tpid=$rows{TAPEID};
					}

				# the label is good, go on...
				$currtime = localtime (time ());
				print ("[$$] $currtime starting tape operations, tppos $tapeposition.\n") if ($VERBOSE);


				################################################

				#########################
				## if the job is to save, or a differential save
				JOBSAVE :
				if (($rows{JOBTYPE} eq 'SAVE') or ($rows{JOBTYPE} eq 'DIFFSAVE') )
					{
					my ($err, $res, $sth, @archrows, $what, $archid, $archpos, $archdesc, @res);
					my $archset='';
					if (!$newtape)
						{
						# goto EOF
						($err, $res)=&GotoTapeEnd($DriveVar[$i]{'dr_dev'}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						if ($err)
							{
							&ExitOnError($WarningMes.="following error occured at End of data : $err", $rows{JOBID});
							}
						}
					# $rows{WHAT} contains list of file to save if SAVE, or JOBID from which get WHAT if DIFFSAVE
					$what=$rows{WHAT};

					# grab info for DIFFSAVE
					if ($rows{JOBTYPE} eq 'DIFFSAVE')
						{
						# get what to save and referring archive
						$request="SELECT WHAT, ARCHID FROM alljobs WHERE JOBID='$rows{WHAT}'";
						print "[$$] Request : $request \n\n" if ($DEBUG);
						$sth=$dbhdle->prepare($request);
						$sth->execute;
						($what, $archset)=$sth->fetchrow_array or &ExitOnError($WarningMes.="[$$] FATAL : no result from DB for DIFFSAVE!\n", $rows{JOBID});
						# get archset info from referring archive
						$request="SELECT ARCHSET FROM allarchives WHERE ARCHID='$archset'";
						print "[$$] Request : $request \n\n" if ($DEBUG);
						$sth=$dbhdle->prepare($request);
						$sth->execute;
						my ($prevset)=$sth->fetchrow_array or warn("No Archset info available.\n");
						$archset=$prevset.$archset if ($prevset);
						$archset.=":";
						}
					# exit if nothing to save...
					&ExitOnError($WarningMes.="[$$] FATAL : Nothing to SAVE!\n", $rows{JOBID}) if (!$what);

					## create an archive
					# get biggest archid, create biggest+1
					$request="SELECT ARCHID FROM allarchives WHERE 1 ORDER BY ARCHID DESC LIMIT 1";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->prepare($request);
					$sth->execute;
					@archrows=$sth->fetchrow_array or ($archrows[0]=0);
					$sth->finish;
					$archid=$archrows[0]+1;
					$rows{ARCHID}=$archid;
					$archpos=$tapeposition+1;

					# insert basic values : archid, intape $tplabel, posontp, creator USER, ARCHNAME, DESCR
					$request="insert into allarchives (ARCHID, TAPEID, POSONTP, CREATOR, ARDTTIME, ARCHSET,  ARCHNAME, DESCR, WHAT ) values (";
					#compute current time for SQL...
					$SQLtime=strftime "%Y-%m-%d %H:%M:%S", localtime;
					$rows{ARCHNAME}=&DBquote($rows{ARCHNAME}) or $rows{ARCHNAME}='No name';
					$archdesc=&DBquote($rows{DESCR}) or $archdesc='';
					$request.="$archid, $tpid, $archpos, '$rows{CREATOR}', '$SQLtime', '$archset', '$rows{ARCHNAME}', '$archdesc', '$what')";
					$sth=$dbhdle->do($request) or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't create archive $archid ! request : $request\n", $rows{JOBID});
					print "[$$] Request : $request \n\n" if ($DEBUG);
					my $archname='ARCHID'.$archid;
					my $archdir=$DriveVar[$i]{'dr_dev'}.'/ARCHID'.$archid;

					if ($DriveVar[$i]{'dr_fam'} eq 'streamer')
						{
						# create an "archive header file" and save it as a separate archive :
						open (NEWARCH, ">/var/tmp/$archname") or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't create archive header $archname'! !\n", $rows{JOBID}) ;
						print NEWARCH ("$archname\nARCHSET:$archset\nTAPEPOS:$tapeposition\nCREATOR:$rows{CREATOR}\nARDTTIME:$SQLtime\nARCHNAME:$rows{ARCHNAME}\nDESCR:$rows{DESCR}\n");
						close (NEWARCH);

						# and save it
						($err, @res)=&PutToArchive($DriveVar[$i]{'dr_dev'}, '/var/tmp', $archname,$rows{JOBID});
						# then delete it
						unlink "/var/tmp/$archname";
						&ExitOnError ( $WarningMes.="[$$] FATAL : Can't save archive header $archname'! error $err !\n", $rows{JOBID}) if ($err);
						}
					else
						{
						# create a folder to put file inside...
						mkdir ($archdir, 0777) or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't create archive folder $archdir'! error $err !\n", $rows{JOBID});
						# here we should someday save archive info too...
						}
					## save files
					@TheFiles=split(/\n/, $what);
					my $FileSize=0;
					foreach (@TheFiles)
						{
						my $file=$_;
						$file=~ s/\r//g;
						my ($res, $filecksum, $filenm);
						if (-d $file)
							{
							# run veracity -THIS IS THE ONLY GUARANTEE THAT ARCHIVE IS NOT EMPTY
							($err, @res)=&GenDirCkSum($file);
							$WarningMes.="[$$] WARNING : Following error generating cksum: $err\n" if ($err);
							}

						#get file size
						($err, $res)=&GetFileSize("$file");
						$FileSize+=$res if ($res);
						$WarningMes.="[$$] WARNING : Following error getting file size: $err\n" if ($err);
						print "[$$] Archive Size $FileSize KB\n"if ($DEBUG);
						# Compare to space available on tape!!!!
						$WarningMes.="[$$] WARNING : space available $tpspace MB, archive size ".int($FileSize/1024)." MB!\n" if (($FileSize/1024)>$tpspace);

						# save the file if it exists!
						if (-e $file)
							{
							my ($fileerr, @gosave);
							if ($DriveVar[$i]{'dr_fam'} eq 'streamer')
								{
								($err, @gosave)=&PutToArchive($DriveVar[$i]{'dr_dev'}, "/", $file,$rows{JOBID}, $dbhdle, $tpid, $archid) ;
								}
							elsif ($DriveVar[$i]{'dr_fam'} eq 'fixeddisk')
								{
								($fileerr, @gosave)=&CopyToDir($archdir, "/", $file, $dbhdle, $tpid, $archid, $rows{JOBID}) ;
								$err.=$fileerr;
								}
							push(@workedFiles, @gosave) if (@gosave);
							}
						$WarningMes.="[$$] WARNING : Following error occured while saving files : $err\n" if ($err);
						# then remove checksum files.
						my @chkfiles=split (/\|/, $ConfVar{'dirchk_files'});
						foreach (@chkfiles)
							{
							my $delfile=$file.'/'.$_;
							if (-d $file)
								{
								unlink ($delfile);
								print "[$$] delete chk file $delfile\n" if ($DEBUG);
								}
							}
						}


					# update archive date and size
					$SQLtime=strftime "%Y-%m-%d %H:%M:%S", localtime;
					$request="UPDATE allarchives set ARENDDTTIME='$SQLtime', ARSIZEK=$FileSize where ARCHID=$archid";
					$sth=$dbhdle->do($request) or ( $WarningMes.="[$$] WARNING : Can't update arch. $archid end time to '$SQLtime' and size to '$FileSize'! !\n");

					# create end mark on tape...
					($err, $res)=&CreateEndMark($DriveVar[$i]{'dr_dev'}, $tpid) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');

					# update remaining Space on tape
					$tpspace-=int($FileSize/1024)+1;
					$request="UPDATE alltapes SET TPSPACE='$tpspace' WHERE TAPEID='$tpid'";
					if ($newtape)
						{
						$SQLtime=strftime "%Y-%m-%d %H:%M:%S", localtime;
						$request="REPLACE INTO alltapes ( TAPEID, TPLABEL, TPDATETIME, TPSPACE, TPFORMAT) values (";
						$request.=" $tpid, '$tplabel', '$SQLtime', $tpspace, '$DriveVar[$i]{'tp_type'}' )";
						}
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->do($request) or $WarningMes.="[$$] WARNING : Can't update tape size for tape $tpid! \n Size should be '$tpspace'!\n";
					}


				###############################
				# if the job is to restore...
				JOBRESTORE :
				if ($rows{JOBTYPE} eq 'RESTORE')
					{
					# ARCHID is mandatory!
					&ExitOnError ( $WarningMes.="[$$] FATAL : No archive defined! job failed\n", $rows{JOBID}) if ($rows{ARCHID} eq '');
					# job global variables
					my ($err,$res, @listed, @posrows);

					# get pos on tape...
					$request="SELECT POSONTP, ARSIZEK FROM allarchives WHERE ARCHID=$rows{ARCHID}";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->prepare($request);
					$sth->execute;

					@posrows=$sth->fetchrow_array or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't get Archive position from DB!: $request failed\n", $rows{JOBID});
					$sth->finish;
					# transport tape
					if (!$tapeposition)
						{
						($err, $res)=&RewindTape($DriveVar[$i]{'dr_dev'}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						$WarningMes.="Following error occured : $err" if ($err);
						# forward according to :
						# pos 0 : TAPEID pos 1: ARCHID1 pos 2: ARCH 1 pos 3: ARCHID2 pos 4 ARCH2
						($err,$res)=&ForwardTape($DriveVar[$i]{'dr_dev'}, $posrows[0]-1) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						}
					else
						{
						($err, $res)=&GotoTapePos($DriveVar[$i]{'dr_dev'},$posrows[0]) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						}
					$WarningMes.="Following error occured : $err" if ($err);

					# check that's the right place
					if ($DriveVar[$i]{'dr_fam'} eq 'streamer')
						{
						($err, @listed)=&GetArchId($DriveVar[$i]{'dr_dev'});
						print "[$$] Pos on tape : $listed[0], tppos $tapeposition\n" if ($DEBUG);
						$WarningMes.=$err if ($err);
						chomp $listed[0];
						if ($listed[0] ne "ARCHID$rows{ARCHID}")
							{
							$tapeposition=-1;
							&ExitOnError( $WarningMes.="[$$] FATAL : Problem on tape : waiting for ARCHID$rows{ARCHID}, got $listed[0]", $rows{JOBID} );
							}
						else
							{
							# forward to the next record (beginning of actual archive)
							($err, $res)=&ForwardTape($DriveVar[$i]{'dr_dev'}, 1) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
							$WarningMes.="Following error occured : $err" if ($err);
							}
						}

					# Get available space on disk...
					# $posrows[1] = espace ncessaire en Ko slack compris
					# on considre l'espace si un dossier cible est dtermin!
					my $avspace;
					($err, $avspace)=&GetDriveSpace($rows{FROMDIR});
					if ($posrows[1]>$avspace)
						{
						$WarningMes.="[$$] WARNING : Not enough room on target filesystem : $posrows[1] KB required, $avspace KB available.\n";
						}
					if ($err)
						{
						$WarningMes.="[$$] WARNING : Couldn't get available space on target Filesystem.\n";
						}

					# restore files
					if ((!$rows{WHAT}) or ($rows{ISARCHIVE} eq 'Y'))# extract the whole archive
						{
						push @TheFiles, "*";
						}
					else
						{
						@TheFiles=split(/\n/, $rows{WHAT});
						}

					foreach (@TheFiles)
						{
						# here we should go on until we reach the next "ARCHID" or "TAPEEND" record.
						my $stop;
						while (!$stop)
							{
							my $file=$_;
							my @golist;
							($err, @golist)=&RestFromArchive($DriveVar[$i]{'dr_dev'}, $rows{FROMDIR}, $file, $rows{JOBID}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
							($err, @golist)=&CopyToDir($rows{FROMDIR},$DriveVar[$i]{'dr_dev'}.'/ARCHID'.$rows{ARCHID},$file,$rows{JOBID}) if ($DriveVar[$i]{'dr_fam'} eq 'fixeddisk');
							print "[$$] Restoring @golist \n" if ($VERBOSE);
							$WarningMes.="[$$] WARNING : the following error occured while restoring : $err\n" if ($err);
							$stop++ if ($err); #end of tape or other error.
							$stop++ if (join("", @golist)=~ m/ARCHID|TAPEEND/);#reached end of archive or end of tape
							if (@golist and !$stop)
								{
								push (@workedFiles, @golist);
								}
							$stop++  if ($DriveVar[$i]{'dr_fam'} ne 'streamer'); # no need to repeat if it's not a streamer.
							}
						# Then we go back to the beginning of archive.
						($err, $res)=&GotoTapePos($DriveVar[$i]{'dr_dev'}, $posrows[0]) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						$WarningMes.="[$$] WARNING : the following error occured during tape transport : $err\n" if ($err);
						}
					}

				###############################
				# if the job is to compare files...
				JOBCOMPARE:
				if ($rows{JOBTYPE} eq 'COMPARE')
					{
					# restore chkfiles for archive ARCHID
					# ARCHID is mandatory!
					&ExitOnError ( $WarningMes.="[$$] FATAL : No archive defined! job failed\n", $rows{JOBID}) if ($rows{ARCHID} eq '');
					# job global variables
					my ($err,$res, @listed, @posrows);

					# get pos on tape and list of files to check...
					$request="SELECT POSONTP, WHAT FROM allarchives WHERE ARCHID=$rows{ARCHID}";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->prepare($request);
					$sth->execute;

					@posrows=$sth->fetchrow_array or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't get Archive position and file list from DB!: $request failed\n", $rows{JOBID});
					$sth->finish;
					# transport tape
					if (!$tapeposition)
						{
						($err, $res)=&RewindTape($DriveVar[$i]{'dr_dev'}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						$WarningMes.="Following error occured : $err" if ($err);
						# forward according to :
						# pos 0 : TAPEID pos 1: ARCHID1 pos 2: ARCH 1 pos 3: ARCHID2 pos 4 ARCH2
						($err,$res)=&ForwardTape($DriveVar[$i]{'dr_dev'}, $posrows[0]-1) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						}
					else
						{
						($err, $res)=&GotoTapePos($DriveVar[$i]{'dr_dev'},$posrows[0]) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						}
					$WarningMes.="Following error occured : $err" if ($err);

					# check that's the right place
					if ($DriveVar[$i]{'dr_fam'} eq 'streamer')
						{
						($err, @listed)=&GetArchId($DriveVar[$i]{'dr_dev'});
						print "[$$] Pos on tape : $listed[0], tppos $tapeposition\n" if ($DEBUG);
						$WarningMes.=$err if ($err);
						chomp $listed[0];
						if ($listed[0] ne "ARCHID$rows{ARCHID}")
							{
							$tapeposition=-1;
							&ExitOnError( $WarningMes.="[$$] FATAL : Problem on tape : waiting for ARCHID$rows{ARCHID}, got $listed[0]", $rows{JOBID} );
							}
						else
							{
							# forward to the next record (beginning of actual archive)
							($err, $res)=&ForwardTape($DriveVar[$i]{'dr_dev'}, 1) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
							$WarningMes.="Following error occured : $err" if ($err);
							}
						}
					# What to check?
					# it should be possible to extract this info directly from tape, but...
					@TheFiles=split(/\n/, $posrows[1]);
					print "[$$] source: $posrows[1], dest: @TheFiles\n" if ($DEBUG);
					my @chkfiles=split (/\|/, $ConfVar{'dirchk_files'});
					my @TheChkFiles; # this will be the ACTUAL list of files to extract
					foreach (@TheFiles)
						{
						chomp;
						my $dir=$_;
						# we extract only chkcsum for directories that EXIST. if it doesn't exist of course it's changed...
						if (-d $dir)
							{
							foreach (@chkfiles)
								{
								push(@TheChkFiles, $dir.'/'.$_);
								}
							}
						}
					# DEBUG
					print "[$$] COMPARE : checkfiles list \n @TheChkFiles\n" if ($DEBUG);
					foreach (@TheChkFiles)
						{
						# here we should go on until we reach the next "ARCHID" or "TAPEEND" record.
						my $stop;
						while (!$stop)
							{
							my $file=$_;
							my @golist;
							($err, @golist)=&RestFromArchive($DriveVar[$i]{'dr_dev'}, '/', $file, $rows{JOBID}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
							($err, @golist)=&CopyToDir('/',$DriveVar[$i]{'dr_dev'}.'/ARCHID'.$rows{ARCHID},$file, $rows{JOBID}) if ($DriveVar[$i]{'dr_fam'} eq 'fixeddisk');
							print "[$$] Compare Restoring @golist \n" if ($VERBOSE);
							$WarningMes.="[$$] WARNING : the following error occured while restoring chksums: $err\n" if ($err);
							$stop++ if ($err); #end of tape or other error.
							$stop++ if (join("", @golist)=~ m/ARCHID|TAPEEND/);#reached end of archive or end of tape
							print "[$$] restored checksum file: @golist\n" if ($DEBUG);
							$stop++  if ($DriveVar[$i]{'dr_fam'} ne 'streamer'); # no need to repeat if it's not a streamer.
							}
						# Then we go back to the beginning of archive.
						($err, $res)=&GotoTapePos($DriveVar[$i]{'dr_dev'}, $posrows[0]) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						$WarningMes.="[$$] WARNING : the following error occured during tape transport : $err\n" if ($err);
						}
					}



				###############################
				# if the job is to compare or verify files...
				JOBVERIFY:
				if (($rows{JOBTYPE} eq 'COMPARE') or ($rows{JOBTYPE} eq 'VERIFY'))
					{
					# ARCHID is mandatory!
					&ExitOnError ( $WarningMes.="[$$] FATAL : No archive defined! job failed\n", $rows{JOBID}) if ($rows{ARCHID} eq '');
					# job global variables
					my ($err,$res, @changedfiles, @files);

					# get list of files to check...
					$request="SELECT WHAT FROM allarchives WHERE ARCHID=$rows{ARCHID}";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->prepare($request);
					$sth->execute;

					($res)=$sth->fetchrow_array or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't get Archive position from DB!: $request failed\n", $rows{JOBID});
					$sth->finish;
					# control checkcsums, Set list of changed files.
					@files=split(/\n/,$res);
					foreach (@files)
						{
						chomp;
						print "[$$] file to verify:'$_'\n" if ($DEBUG);;
						my $ischanged=0;
						if (-d $_)
							{
							# it's a dir, control dir checksum
							my (@filelist);
							($err, @filelist)=&CtrlDirCkSum("/$_");
							$WarningMes.="following error occured while dir chksum: $err\n" if ($err);
							print "[$$] CtrlDirCkSum returned :\n@filelist\n" if ($DEBUG);
							push(@changedfiles, @filelist);
							push(@workedFiles, join(" CHANGED\n", @filelist)." CHANGED\n");
							}
						elsif (-e $_)
							{
							my ($dbchksum, $filecksum);
							# it's a file, get chksum from DB
							# what's the dir the file's in?
							my @dirs=split('/', $_);
							my $eachfile=pop @dirs;
							my $dir=join("/", @dirs);
							# initial / aren't stored in allfiles table
							$dir=~ s#^/##;;
							$request="SELECT MD5SUM FROM allfiles WHERE ( FILEPATH='$dir' AND FILENAME='$eachfile' AND ARCHID='$rows{ARCHID}' )";
							print "[$$] Request : $request \n\n" if ($DEBUG);
							$sth=$dbhdle->prepare($request);
							$sth->execute;
							($dbchksum)=$sth->fetchrow_array or ($WarningMes.="[$$] WARNING : no result from DB for query $request\n");
							($err, $filecksum)=&FileCkSum("/$_");
							if ($err)
								{
								$WarningMes.="/$_ not a regular file, no action!!\n";
								}
							elsif ($filecksum ne $dbchksum)
								{
								print "[$$] Verify chksums don't match, DB :$dbchksum, file : $filecksum\n" if ($DEBUG);
								$ischanged=1;
								}
							else
								{
								print "[$$] Verify chksums match, DB :$dbchksum, file : $filecksum\n" if ($DEBUG);
								$ischanged=0;
								push(@workedFiles, $_.' identical');
								}

							}
						else
							{
							# it's been deleted or moved
							$ischanged=1;
							}
						if ($ischanged)
							{
							push(@changedfiles, $_);
							push(@workedFiles, $_.' CHANGED');
							}
						}

					# write list in the WHAT field for the job
					if (@changedfiles)
						{
						$request="UPDATE alljobs SET WHAT='".&DBquote(join("\n", @changedfiles))."' WHERE JOBID=$rows{JOBID}";
						my $sth=$dbhdle->do($request) or &ExitOnError( $WarningMes.="[$$] ERROR : Can't update job $rows{JOBID} changed files list!\nList following:\n@changedfiles\n", $rows{JOBID});
						}
					}


				###############################
				# if the job is to list  files...
				JOBLIST :
				if ($rows{JOBTYPE} eq 'LIST')
					{
					my ($err,$res, @listed, @posrows);
					# ARCHID is mandatory!
					&ExitOnError ( $WarningMes.="[$$] FATAL : No archive defined! job failed\n", $rows{JOBID}) if ($rows{ARCHID} eq '');

					# get pos on tape...
					$request="SELECT POSONTP, ARSIZEK FROM allarchives WHERE ARCHID=$rows{ARCHID}";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->prepare($request);
					$sth->execute;

					@posrows=$sth->fetchrow_array or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't get Archive position from DB!: $request failed\n", $rows{JOBID});
					$sth->finish;
					# transport tape
					if (!$tapeposition)
						{
						($err, $res)=&RewindTape($DriveVar[$i]{'dr_dev'}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						$WarningMes.="Following error occured : $err" if ($err);
						# forward according to :
						# pos 0 : TAPEID pos 1: ARCHID1 pos 2: ARCH 1 pos 3: ARCHID2 pos 4 ARCH2
						($err,$res)=&ForwardTape($DriveVar[$i]{'dr_dev'}, $posrows[0]-1) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						}
					else
						{
						($err, $res)=&GotoTapePos($DriveVar[$i]{'dr_dev'},$posrows[0]) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						}
					$WarningMes.="Following error occured : $err" if ($err);

					# check that's the right place
					($err, @listed)=&GetArchId($DriveVar[$i]{'dr_dev'}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
					print "[$$] Pos on tape : $listed[0], tppos $tapeposition\n" if ($DEBUG);
					$WarningMes.=$err if ($err);
					chomp $listed[0];
					if ($listed[0] ne "ARCHID$rows{ARCHID}")
						{
						$tapeposition=-1;
						&ExitOnError( $WarningMes.="[$$] FATAL : Problem on tape : waiting for ARCHID$rows{ARCHID}, got $listed[0]", $rows{JOBID} );
						}
					else
						{
						# forward to the next record (beginning of actual archive)
						($err, $res)=&ForwardTape($DriveVar[$i]{'dr_dev'}, 1) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						$WarningMes.="Following error occured : $err" if ($err);
						}

					# list files
					if ($rows{ISARCHIVE} eq 'Y')
						{
						push @TheFiles, " ";
						}
					else
						{
						@TheFiles=split(/\n/, $rows{WHAT});
						}

					foreach (@TheFiles)
						{
						# here we should go on until we reach the next "ARCHID" or "TAPEEND" record.
						my $stop;
						while (!$stop)
							{
							my $file=&DBquote($_);
							my @golist;
							($err, @golist)=&ListFromArchive($DriveVar[$i]{'dr_dev'}, $rows{JOBID}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
							print "[$$] Listing @golist \n" if ($VERBOSE);
							$WarningMes.="[$$] WARNING : the following error occured while reading tape : $err\n" if ($err);
							$stop++ if ($err);
							$stop++ if (join("", @golist)=~ m/ARCHID|TAPEEND/);
							push (@workedFiles, @golist) unless ($stop);
							}
						# Then we go back to the beginning of archive.
						($err, $res)=&GotoTapePos($DriveVar[$i]{'dr_dev'}, $posrows[0]) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
						$WarningMes.="[$$] WARNING : the following error occured during tape transport : $err\n" if ($err);
						}
					}

				 ###############################
				# if the job is a purge job...
				JOBPURGE :
				if ($rows{JOBTYPE} eq 'PURGE')
					{
					&ExitOnError ( $WarningMes.="[$$] FATAL : No archive defined! job failed\n", $rows{JOBID}) if ($rows{ARCHID} eq '');
					# job global variables
					my ($err,$res, $purge);
					
					# check the archive from DB...
					$request="SELECT ARCHID FROM allarchives WHERE ARCHID=$rows{ARCHID}";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->prepare($request);
					$sth->execute;

					($res)=$sth->fetchrow_array or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't get Archive info from DB!: $request failed\n", $rows{JOBID});
					$sth->finish;
					
					# It's OK, remove everything!
					$purge=&purgeArchive ( $dbhdle, $DriveVar[$i]{'dr_dev'}, $res );
					$WarningMes.="[$$] WARNING : the following errors occured while purging Archive '$res' :\n$purge" if ($purge);
					}
				
				 ###############################
				# if the job is a migration job...
				JOBMIGRATE :
				if ($rows{JOBTYPE} eq 'MIGRATE')
					{
					&ExitOnError ( $WarningMes.="[$$] FATAL : No archive defined! job failed\n", $rows{JOBID}) if ($rows{ARCHID} eq '');
					# job global variables
					my ($err, $restape, $resdrv, $respath, $desttape);
					
					# check the archive from DB...
					$request="SELECT TAPEID FROM allarchives WHERE ARCHID=$rows{ARCHID}";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->prepare($request);
					$sth->execute;
					($restape)=$sth->fetchrow_array or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't get Archive info from DB!: $request failed\n", $rows{JOBID});
					$sth->finish;
					
					# get the drive from allslots
					$request="SELECT DRIVEID FROM allslots WHERE TAPEID=$restape AND TAPEIN='Y'";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->prepare($request);
					$sth->execute;

					($resdrv)=$sth->fetchrow_array or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't get Drive info from DB!: $request failed\n", $rows{JOBID});
					$sth->finish;
					
					# get the path from alldrives (we need better sql...)
					$request="SELECT DRDEV FROM alldrives WHERE DRIVEID=$resdrv";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->prepare($request);
					$sth->execute;
					
					($respath)=$sth->fetchrow_array or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't get respath  from DB!: $request failed\n", $rows{JOBID});
					# last get the new tape the data will migrate to...
					$request="SELECT TAPEID FROM allslots WHERE DRIVEID=$rows{DRIVEID}";
					print "[$$] Request : $request \n\n" if ($DEBUG);
					$sth=$dbhdle->prepare($request);
					$sth->execute;
					
					($desttape)=$sth->fetchrow_array or &ExitOnError ( $WarningMes.="[$$] FATAL : Can't get dest. tapeid from DB!: $request failed\n", $rows{JOBID});
					$sth->finish;
					# first move data
					# source folder
					my $archpath=$respath.'/ARCHID'.$rows{ARCHID} ;
					print "[$$] source : $archpath\n" if ($DEBUG);	
					if ( -e $archpath ) {
						# we should use a transaction for that instead
						# copy data
						my ($cperror, @result)=&CopyToDir($DriveVar[$i]{'dr_dev'}, $respath, 'ARCHID'.$rows{ARCHID});
						warn ("[$$] ERROR : $cperror ") if ($cperror);
						print "[$$] copy:".join("\n", @result)."\n" if ($DEBUG);

						# then check it's well done
						# TODO
						unless ($cperror) {
							# update the database values
							my $request="UPDATE allfiles set TAPEID='$desttape' where ARCHID=$rows{ARCHID}";
							my $sth=$dbhdle->do($request) or ( warn ( "[$$] WARNING : Can't update files in DB! request : $request \n"));
							print "Request : $request \n\n" if ($DEBUG);
							$request="UPDATE allarchives set TAPEID='$desttape' where ARCHID=$rows{ARCHID}";
							$sth=$dbhdle->do($request) or ( warn ("[$$] WARNING : Can't update archive $rows{ARCHID} from DB! request : $request \n"));
							print "Request : $request \n\n" if ($DEBUG);

							# It's OK, remove everything!
							my $command="rm -rf $archpath";
							#print "[$$] issuing \"$command\" \n" if ($VERBOSE);
							#my $pid=open3 (\*WRH,\*RDH, \*ERH, $command ) or warn ("[$$] ERROR : couldn't run command : $command !");
							my ($rmerror,$result)=&RunExtCmd($command, "removeArchive");
							warn ("[$$] ERROR : $rmerror !") if ($cperror);

							}
						else
							{
							warn "[$$] ERROR: files not removed from '$archpath'!\n";
							}
						}
					else	{
						# we should extract the data to some temporary space
						# and migrate it as we can
						# for now we just fail gracefully... :(
						&ExitOnError( $WarningMes.="[$$] FATAL : migration impossible (missing feature!\n", $rows{JOBID});
						}
					}
					
				###############################
				# if the job is a test job...
				JOBTEST :
				if ($rows{JOBTYPE} eq 'TEST')
					{
					my ($tape, $res, $err, @golist);
					# log everything
					print "main variables status : ".join(":", %rows)." \n";

					# Check commands
					#Goto start
					($err, $tape)=&GetTapeID($DriveVar[$i]{'dr_dev'}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
					print "[$$] TAPE $tape\n" if ($VERBOSE);
					push @workedFiles, $tape;

					#Goto endmark
					($err, $res)=&GotoTapeEnd($DriveVar[$i]{'dr_dev'}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
					print "[$$] END $err, $res\n" if ($VERBOSE);

					#Goto somewhere in themiddle...
					my  $where=int($res/2);

					($err, $tape)=GotoTapePos($DriveVar[$i]{'dr_dev'}, $where) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
					print "[$$] GOTO  $where, get  $tape\n" if ($VERBOSE);

					#list  : get archive
					($err, @golist)=&ListFromArchive($DriveVar[$i]{'dr_dev'}, $rows{JOBID}) if ($DriveVar[$i]{'dr_fam'} eq 'streamer');
					print "[$$] Listing @golist \n" if ($VERBOSE);
					push @workedFiles, @golist;

					}

				#########################
				# end of job
				#########################
				# mark job as done
				CLOSEJOB :
					{
					my $Message=join "\n", @workedFiles;
					if ($WarningMes)
						{
						open (JOBLOG, '>>', "$homedir/../logs/job".$rows{JOBID}.".log") or warn ("Can't open job $rows{JOBID} log for writing: $!");
						print JOBLOG $WarningMes;
						close JOBLOG;
						$currtime = localtime (time ());
						warn ("$currtime JOB $rows{JOBID} finished with errors. \n$WarningMes \n");
						$Message.="\nError messages\n".$WarningMes;
						$WarningMes='';
						if (($rows{REPORTTO}) and ($rows{REPORTTO}!~ m/ADMIN/ ))
							{
							$rows{REPORTTO}.="\nADMIN\n";
							}
						elsif (!$rows{REPORTTO})
							{
							$rows{REPORTTO}.="ADMIN\n";
							}
						}
					

					$Message=&DBquote($Message);
					$SQLtime=strftime "%Y-%m-%d %H:%M:%S", localtime;
					$request="UPDATE alljobs set STATUS='DONE', MESSAGE='$Message', JENDTIME='$SQLtime', ARCHID='$rows{ARCHID}' where JOBID=$rows{JOBID}"; # insrer la liste dans le champ "message"
					my $sth=$dbhdle->do($request) or &ExitOnError ( $WarningMes.="[$$] ERROR : Can't update job $rows{JOBID} status to DONE! !\n", $rows{JOBID});
					print "[$$] Request : $request \n\n" if ($DEBUG);

					$currtime = localtime (time ());
					print "$currtime $rows{JOBTYPE} job $rows{JOBID}, done\n\n";
					
					# send emails
					if ($rows{REPORTTO})
						{
						# split all emails
						my @themails=split(/\n/, $rows{REPORTTO});
						foreach my $themail (@themails)
							{
							$themail=$ConfVar{'adm_mail'} if ($themail eq 'ADMIN');
							#should check emails are valid someday.
							# SndEmail (destmail, destname, subject, body);
							&SndEmail($themail,$_,"StorIQ archiver $rows{JOBTYPE} job $rows{JOBID} complete","$rows{JOBTYPE} Job $rows{JOBID}, with description :\n\"$rows{DESCR}\"\n was completed at $currtime.\n Files :\n$Message\n");
							}
						}
					}
				}
			else
				{
				print "[$$] Nothing to do, snoozing a while...\n" if ($VERBOSE);
				$term=1 if ($EXITWHENDONE);
				sleep(60) if (!$term);
				}
			$term=1 if ($INITONLY);
			}
		}
	elsif ($pid>0)
		{
		push (@processlist, $pid);
		print "process $pid running for drive $i \n" if ($DEBUG);
		}
	}

#############################
# before exiting, it would be fine to clean the place  :
# unload tapes
# check that no running jobs are left behind.

# declare here, for the father only...
$SIG{CHLD} =\&REAPER;

# father's nursing
while (!$term)
	{
	sleep 10;


	if ($child)
		{
		print "waiting for child to be killed.\n";
		my $wpid=wait;
		print "... [$wpid] terminated.\n";
		$term=1;
		}
	}

foreach my $tmp(@processlist)
	{
	print "explicit kill [$tmp]...\n";
	kill('TERM', $tmp);
	}


# end of program so far
if (@processlist)
	{
	# wait until childs are closed.
	sleep 10;
	$currtime = localtime (time ());
	warn ("$currtime StorIQ Archiver V $Version exiting...\n");
	exit;
	}
else
	{
	print "[$$] ...process closed\n";
	exit;
	}

# never executed:
die "abnormal program termination\n";

#################################
# tape management functions
#################################
# initialize a New sTape and label it
# dr dev and tape id
sub NewTape
	{
	my ($dev, $tpid)=@_;
	&RewindTape($dev);
	my $filename='TAPEID'.$tpid;
	open (NEWTAPE, ">/var/tmp/$filename");
	print NEWTAPE ("$filename") or warn "[$$] WARNING : can't write TAPE HEADER!\n";        # or tar may not save...
	close (NEWTAPE);
	# do not hesitate, erase the tape!
	&EraseTape($dev);
	&RewindTape($dev);
	#save the label to tape
	my $command="cd /var/tmp; $ConfVar{'TAR_prg'} cf $dev $filename";
	my  ($error,@result)=&RunExtCmd($command, "NewTape");
	unlink "/var/tmp/$filename";
	$tapeposition=1;
	# returns happy boy
	return ($error, $tapeposition);
	}

# get TapeID from tape itself
# returns -1 if unused; else return the numeric value.
sub GetTapeID
	{
	my $dev=$_[0];
	my ($err, $result)=&RewindTape($dev);
	($err, $result)=&GetArchId($dev);
	# if the tape is OK, the first archive contains TAPEID<tapenumber>
	if ($result)
		{
		print "[$$] FOUND ON TAPE $result \n" if ($VERBOSE);
		my ($tp, $tpid)=split(/ID/, $result);
		if ($tp eq 'TAPE')
			{
			$result=$tpid;
			}
		else
			{
			# alien tape; erase for now
			$result=-1;
			}

		}
	else
		{
		#new tape
		$result=-1;
		}
	return ($err, $result);
	}

# create an endmark to where the tape is...
sub CreateEndMark
	{
	my ($dev, $tpid)=@_;
	my $filename='TAPEEND'.$tpid; my $pos=$tapeposition+1;
	open (ENDTAPE, ">/var/tmp/$filename");
	print ENDTAPE ("$filename\n$pos\n");
	close (ENDTAPE);
	#save the label to tape
	my $command="cd /var/tmp; $ConfVar{'TAR_prg'} cf $dev $filename";
	my ($error,@result)=&RunExtCmd($command, "CreateEndMark");
	$tapeposition+=1 if (@result);
	print "CreateEndMark wr pos $pos tppos : $tapeposition\n" if ($DEBUG);
	unlink "/var/tmp/$filename";
	return ($error, $tapeposition);
	}

# rewind tape to the beginning
sub RewindTape
	{
	my $dev=$_[0];
	my $command="$ConfVar{'MT_prg'} -f $dev rewind";
	my ($error,@result)=&RunExtCmd($command, "RewindTape");
	$tapeposition=0 if (@result);
	print "Rewind tppos : $tapeposition\n" if ($DEBUG);
	return ( $error, @result);
	}


#########################
# move to position $dest according to current pos...
#

sub GotoTapePos
	{
	my ($dev, $dest)=@_;
	my ($err, $res);
	if ($dest>$tapeposition)
		{
		($err, $res)=&ForwardTape($dev, $dest-$tapeposition);
		}
	elsif ($dest<$tapeposition)
		{
		($err, $res)=&BackwardTape($dev, $tapeposition-$dest);
		}
	print "GotoTapePos tppos : $tapeposition\n" if ($DEBUG);
	return ($err, $tapeposition);
	}


# goto end of tape data
# params : device
# returns tape number found
sub GotoTapeEnd
	{
	# move to  endmark
	my $dev=$_[0]; my @listed;
	my $command="$ConfVar{'MT_prg'} -f $dev eod";
	my ($error,@result)=&RunExtCmd($command, "GotoTapeEnd");
	# get back one record
	my ($err, $res)=&BackwardTape($dev, 2);
	$error.=$err if ($err);
	# check record : should be TAPEEND<tpid>
	($err, @listed)=&RestFromArchive($dev, "/tmp/", "", "ENDTAPE");
	$error.=$err if ($err);
	print "GotoTapeEnd on tape : $listed[0] \n tppos : $tapeposition\n" if ($DEBUG);
	if ($listed[0] !~ m/TAPEEND/)
		{
		# should insert here a reconstruction tool, to count from start and recreate END MARK.
		$error.="[$$] FATAL : Problem on tape : waiting for TAPEEND, got $listed[0]" ;
		}
	else
		{
		# read the position
		my $file=open(TPINFO, "</tmp/$listed[0]");
		my @fileinfo=<TPINFO>;
		close (TPINFO);
		$fileinfo[1]=~ tr/0-9//cd;
		$tapeposition=$fileinfo[1];
		print "tppos : $tapeposition\n";
		unlink("/tmp/$listed[0]");
		# get back one record to overwrite the END mark
		($err, $res)=&BackwardTape($dev, 1);
		$error.=$err if ($err);
		}
	print "GotoTapeEnd tppos : $tapeposition\n" if ($DEBUG);
	return ($error, $tapeposition);
	}


#########################
# Forward tape 
# device, counts
sub ForwardTape
	{
	my ($dev, $cnt)=@_;
	my $command="$ConfVar{'MT_prg'} -f $dev fsf $cnt";
	my ($error,@result)=&RunExtCmd($command, "ForwardTape");
	$tapeposition+=$cnt unless ($error);
	print "ForwardTape tppos : $tapeposition\n" if ($DEBUG);
	return ( $error, @result);
	}

# Rewind tape of some records
sub BackwardTape
	{
	my ($dev, $cnt)=@_;
	my $command="$ConfVar{'MT_prg'} -f $dev bsfm $cnt";
	my ($error,@result)=&RunExtCmd($command, "BackwardTape");
	$tapeposition-=$cnt unless ($error);
	print "BackwardTape tppos : $tapeposition\n" if ($DEBUG);
	return ( $error, @result);
	}

# erase all data on tape
sub EraseTape
	{
	my $dev=$_[0];
	my $command="$ConfVar{'MT_prg'} -f $dev erase";
	my ($error,@result)=&RunExtCmd($command, "EraseTape");
	return ( $error, @result);
	}

# get tape info
sub GetTpInfo
	{
	my $dev=$_[0];
	my $command="$ConfVar{'MT_prg'} -f $dev status";
	my ($error,@result)=&RunExtCmd($command, "GetTpInfo");
	return ( $error, @result);
	}

# changer commands

# get scsi hardware status
# take the scsi device to work with as first parameter
sub GetHwStatus
	{
	my $dev=$_[0];
	my $command="$ConfVar{'MTX_prg'} -f $dev inquiry";
	my ($error,@result)=&RunExtCmd($command, "GetHwStatus");
	return ( $error, @result);
	}

# get changer status : tapes loaded, etc.
# take the scsi device to work with as first parameter
sub GetChgrStatus
	{
	my $dev=$_[0];
	my $command="$ConfVar{'MTX_prg'} -f $dev status";
	my ($error,@result)=&RunExtCmd($command, "GetChgrStatus");
	return ( $error, @result);
	}


# load tape
# parms : scsi device, tape device, streamer (slotnum) to unload, dest. slot
sub LoadTape
	{
	# load the tape...
	my  ($dev, $tpdev, $streamer, $slot)=@_;
	my $command="$ConfVar{'MTX_prg'} -f $dev load $slot $streamer";
	my ($error,@result)=&RunExtCmd($command, "LoadTape");
	
	# then wait for the drive to get ready...
	$command="/bin/mt -f $tpdev status";
	my $wait=30;        # maximum count
	if (!$error)
		{
		while ($wait)
			{
			$wait-=1;
			($error,@result)=&RunExtCmd($command, "LoadTape");
			if ($result[5]=~ m/ONLINE/)
				{
				$wait=0;
				}
			else
				{
				sleep(10);
				$error.="[$$] WARNING : tape drive response timeout, retrying $wait times";
				}
			}
		}
	$tapeposition=0 if (@result);
	return ( $error, join('', @result));
	}


# unload tape
# parms : scsi device, tape device, streamer (slotnum) to unload, dest. slot
sub UnloadTape
	{
	my  ($dev, $tpdev, $streamer, $slot)=@_;
	# first set the drive offline (necessary for some models)
	my $command="$ConfVar{'MT_prg'} -f $tpdev offline";
	my ($error,@result)=&RunExtCmd($command, "UnloadTape");
	
	# then unload the tape if destination defined
	if (!$error)
		{
		my $command="$ConfVar{'MTX_prg'} -f $dev unload $slot $streamer";
		($error,@result)=&RunExtCmd($command, "UnloadTape");
		}
	# workaround for bug in mtx...
	if ($error=~ m/^Unloading/)
		{
		push (@result,$error);
		$error='';
		}

	# return result, error
	$tapeposition=0 if (@result);
	return ($error, join("\n", @result));
	}

# exchange tapes
# ExchTapes (chgdevice, iodevice, fromslot, toslot)
sub ExchTapes
	{
	my  ($dev, $streamer, $fromslot, $toslot )=@_;
	# swap tapes....
	my $command="$ConfVar{'MTX_prg'} -f $dev transfer $fromslot $toslot";
	my ($error,@result)=&RunExtCmd($command, "ExchTapes");
	return ( $error, @result);
	}


#################################
# archive functions
#################################

# write selected files to tape
# tape device to save to, cd to, files to save, job id, [db handle, tape id, arch. id]
sub PutToArchive
	{
	my ($dev, $dir, $file, $jobid, $dbhdle, $tpid, $archid)=@_;
	$dir=~ s/(\W)/\\$1/g;
	$file=~ s#^/+##;
	$file=~ s/(\W)/\\$1/g;
	my $command="cd $dir; $ConfVar{'TAR_prg'} cvf $dev $file";
	print "[$$] func PutToArchive issuing \"$command\" \n" if ($VERBOSE);
	open (ERPH, '>>', "$homedir/../logs/job$jobid.log") or warn ("Can't open job $jobid log for writing: $!");
	my $pid=open3 (\*WRH,\*RDPH, "ERPH", $command ) or warn ("[$$] ERROR : couldn't run command : $command !");
	my (@result,@dberror) ;
	while (<RDPH>)
		{
		chomp;
		push (@result, "/$_");
		my ($dir, $eachfile, $fileparm, @fileattr,$err, $filecksum);
		if ($dbhdle)
			{
			@fileattr=stat("/$_");
			if (-d "/$_")
				{
				$dir=&DBquote("$_");
				$eachfile='';
				$filecksum='dir';
				}
			else
				{
				my @dirs=split('/', $_);
				$eachfile=&DBquote(pop @dirs);
				$dir=&DBquote(join("/", @dirs));
				($err, $filecksum)=&FileCkSum("/$_");
				push(@dberror, $err) if ($err);
				}

			print "[$$] file saved : $dir/$eachfile \n" if ($VERBOSE);
			my $SQLtime=strftime "%Y-%m-%d %H:%M:%S", localtime;
			my $filedate=strftime "%Y-%m-%d %H:%M:%S", localtime($fileattr[9]);
			my $request="insert into allfiles (FILENAME, FILEPATH, FILESIZE, FILEOWNER, FILEDATE, RECDATETIME, TAPEID, ARCHID, MD5SUM) values ('$eachfile', '$dir', '$fileattr[7]', '$fileattr[4]', '$filedate', '$SQLtime', '$tpid', '$archid', '$filecksum' )";
			my $sth=$dbhdle->do($request) or ( push(@dberror, "[$$] WARNING : Can't insert file info in DB! request : $request \n"));
			print "Request : $request \n\n" if ($DEBUG);
			}
		}
	close ERPH;
	open (ERPH, '<', "$homedir/../logs/job$jobid.log") or warn ("Can't open job $jobid for reading: $!");
	my @error=<ERPH>;
	close ERPH;
	push(@error,@dberror);
	waitpid($pid, 0);
	my $sig=($? & 127);
	push (@error, "[$$] WARNING : external command '$command' got signal $sig, errno $!\n") if ( $? & 127 );
	push (@error, "[$$] ERROR : external program CRASHED.\n") if ($? & 128);
	# return result, error
	$tapeposition+=1 if (@result);
	print "PutToArchive tppos : $tapeposition\n" if ($DEBUG);
	return (join("\n",@error), @result);
	}

# restore selected files from archive
# tape device to read from, cd to, files to restore, job id
sub RestFromArchive
	{
	my ($dev, $dir, $file, $jobid)=@_;
	$dir=~ s/(\W)/\\$1/g;
	if ($file ne '*')
		{
		$file=~ s/(\W)/\\$1/g ;
		}
	else
		{
		$file='';
		}
	my $command="cd $dir; $ConfVar{'TAR_prg'} xvf $dev $file";
	my ($error,@result)=&RunExtCmd($command, $jobid);
	# return result, error
	$tapeposition+=1 if (@result);
	return ($error, @result);
	}

# list files from archive
# tape device to read
sub ListFromArchive
	{
	#my $dev;
	my ($dev, $jobid)=@_;
	my $command="$ConfVar{'TAR_prg'} tvf $dev";
	my ($error,@result)=&RunExtCmd($command, $jobid);
	# return result, error
	$tapeposition+=1 if (@result);
	print "ListFromArchive tppos : $tapeposition\n" if ($DEBUG);
	return ($error, @result);
	}

# copy files to a disk
# destination, fromdir, what, database handle, tape, archive...

sub CopyToDir
	{
	my ($dev, $dir, $file, $dbhdle, $tpid, $archid, $jobid)=@_;
	$dir=~ s/(\W)/\\$1/g;
	
	## GROS HACK ##
	my $cmd="$ConfVar{'TAR_prg'} cf -"; # par dfaut, c'est tar
	my $mountpoint;
	my $issmb;
	# sauf si c'est un montage smb:
	open (MTAB, "</etc/mtab");
	while (<MTAB>)
		{
		my (@fsparm)=split (/\s+/);
	$mountpoint=$fsparm[1];
		if (($file =~ m!\Q$mountpoint!) and ($mountpoint ne '/') and  ($fsparm[2]=~ m/^smb/))
			{
			$issmb=$mountpoint;
			# on change la commande
			$cmd="$ConfVar{'SMBCL_prg'} '$fsparm[0]' '$ConfVar{'SMB_password'}' -U '$ConfVar{'SMB_user'}' -Tqc - "; 
			# on tronque le nom du fichier
			$file =~ s!^/?\Q$mountpoint!!;
			# $file='*' if ($file eq '');
			print "[$$] command '$cmd' mountpoint '$mountpoint' target file '$file' \n" if ($DEBUG);
			}
		}
	close (MTAB);
	## GROS HACK ##
	$file=~ s#^/+##;
	$file=~ s/(\W)/\\$1/g if ($file ne '*');
	
	# my $command="cd $dir; $ConfVar{'TAR_prg'} cf - $file |(cd $dev ; $ConfVar{'TAR_prg'} xvf - ) ";
	my $command="cd $dir; $cmd $file |(cd $dev ; $ConfVar{'TAR_prg'} xvf - ) ";
	print "[$$] func CopyToDir issuing \"$command\" \n" if ($VERBOSE);
	open (ERPH, '>>', "$homedir/../logs/job$jobid.log") or warn ("Can't open job $jobid log for writing: $!");
	my $pid=open3 ("</dev/null",\*RDPH, "ERPH", $command ) or warn ("[$$] ERROR : couldn't run command : $command !");
	my (@result,@dberror) ;
	while (my $donefile=<RDPH>)
		{
		chomp $donefile;
		if ($issmb)
			{
			$donefile=~ s#^./##;
			$donefile=$issmb.'/'.$donefile;
			}
		push (@result, "/$donefile");
		my ($dir, $eachfile, $fileparm, @fileattr,$err, $filecksum);
		if ($dbhdle)
			{
			
			if (-d "/$donefile")
				{
				@fileattr=stat("/$donefile") or (push @dberror, "[$$] WARNING : couldn't stat directory '$_'!");
				$dir=&DBquote("$donefile");
				$eachfile='';
				$filecksum='dir';
				}
			else
				{
				if ( -e "/$donefile")
					{
					@fileattr=stat("/$donefile") or (push @dberror, "[$$] WARNING : couldn't stat file '$_'!");
					($err, $filecksum)=&FileCkSum("/$donefile");
					}
				if (( !-e "/$donefile") and ( -l "/$donefile"))
					{
					@fileattr=lstat("/$donefile") or (push @dberror, "[$$] WARNING : couldn't stat broken symlink '$_'!");
					}
				my @dirs=split('/', $donefile);
				$eachfile=&DBquote(pop @dirs);
				$dir=&DBquote(join("/", @dirs));
				push(@dberror, $err) if ($err);
				}

			print "[$$] file saved : $dir/$eachfile \n" if ($VERBOSE);
			my $SQLtime=strftime "%Y-%m-%d %H:%M:%S", localtime;
			my $filedate;
			if ($fileattr[9])
				{
				$filedate=strftime "%Y-%m-%d %H:%M:%S", localtime($fileattr[9]);
				}
			else
				{
				$filedate="00-00-00 00:00:00";
				$fileattr[7]=-1;
				$fileattr[4]="ERROR";
				push @dberror, "[$$] LOURD : couldn't stat file '$donefile'!"
				}
			
			my $request="insert into allfiles (FILENAME, FILEPATH, FILESIZE, FILEOWNER, FILEDATE, RECDATETIME, TAPEID, ARCHID, MD5SUM) values ('$eachfile', '$dir', '$fileattr[7]', '$fileattr[4]', '$filedate', '$SQLtime', '$tpid', '$archid', '$filecksum' )";
			my $sth=$dbhdle->do($request) or ( push(@dberror, "[$$] WARNING : Can't insert file info in DB! request : $request \n"));
			print "Request : $request \n\n" if ($DEBUG);
			}
		}
	close ERPH;
	open (ERPH, '<', "$homedir/../logs/job$jobid.log") or warn ("Can't open job $jobid for reading: $!");
	my $error;
	read (ERPH, $error, 2048) or warn ("can't read log file: $!\n");
	close ERPH;
	$error.=join("\n",@dberror);
	waitpid($pid, 0);
	my $sig=($? & 127);
	$error.="[$$] WARNING : external command '$command' got signal $sig, errno $!\n" if ( $? & 127 );
	$error.="[$$] ERROR : external program CRASHED.\n" if ($? & 128);
	# return result, error
	return ($error, @result);
	}

# simple list (to get IDs from tape)
# GetArchId (iodevice)
sub GetArchId
	{
	#my $dev;
	my $dev=$_[0];
	my $command="$ConfVar{'TAR_prg'} tf $dev";
	my ($error,@result)=&RunExtCmd($command, "GetArchId");
	
	$tapeposition+=1 if (@result);
	print "GetArchId tppos : $tapeposition\n" if ($DEBUG);
	# return result, error
	return ($error, @result);
	}

# create directory checksum
# params: directory

sub GenDirCkSum
	{
	my $dir=$_[0];
	if (!-d $dir)
		{
		return("not a directory!", '');
		}
	my $command="cd $dir; $ConfVar{'dirchk_prg'} snapshot";
	my ($error,@result)=&RunExtCmd($command, "GenDirCkSum");
	
	# return result, error
	my $result="@result";
	if ($result=~ m/success/)
		{
		$result="success";
		}
	else
		{
		$result = '';
		}
	return ($error, $result);
	}

# control checksum
# params : directory
# returns an array of changed files

sub CtrlDirCkSum
	{
	my $dir=$_[0];
	if (!-d $dir)
		{
		return("not a directory!", '');
		}
	my (@changedfiles, $result, $rundir);
	my $command="cd $dir;$ConfVar{'dirchk_prg'} check";
	print "[$$] issuing \"$command\" \n" if ($VERBOSE);
	my $pid=open3 (\*WRH,\*RDCH, \*ERCH, $command ) or warn ("[$$] ERROR : couldn't run command : $command !");
	while ($result=<RDCH>)
		{
		print " dir chksum got \"$result\"\n" if ($DEBUG);
		if ($result=~ /^#/ )
			{
			$rundir=$result;
			$rundir=~ s/^#//g;
			chomp($rundir);
			next;
			}
		print "DIR $rundir, FILE $result\n" if ($DEBUG);
		if ($rundir)
			{
			$result=~ s/^\s+//g;
			chomp($result);
			$result=$dir.$rundir.$result;
			$result=~ s#^/##g;
			print "rebuilt file :$result\n" if ($DEBUG);
			my @splitted=split(/\s+/, $result);
			print "SPLIT ".join(":",@splitted) if ($DEBUG);
			if(pop(@splitted) !~ m/Created./)
				{
				pop(@splitted);
				}
			$result=join(" ",@splitted);
			push(@changedfiles, $result) if ($result);
			}
		}
	my $error;
	read (ERCH, $error, 2048) or warn ("can't read log file: $!\n");
	close ERCH;
	waitpid($pid, 0);
	my $sig=($? & 127);
	$error.="[$$] WARNING : external command '$command' got signal $sig, errno $!\n" if ( $? & 127 );
	$error.="[$$] ERROR : external program CRASHED.\n" if ($? & 128);
	pop(@changedfiles);
	return ($error, @changedfiles);
	}

# single file cksum
# returns MD5 hash for the file (to compare with one stored previously)
# checks only given length (if any)
# FileCkSum (file, length)
sub FileCkSum
	{
	my ($file, $length)=@_;
	my ($data, $sum);
	
	if ((!-e $file) and (-l $file))
		{
		return ("$file : broken symbolic link", "INFO");
		}
	
	if ((!-f $file) and (!-l $file))
		{
		return ("$file : not a regular file", "INFO");
		}
	my $ctx = Digest::MD5->new();
	local *IN;
		open( IN, $file ) || return undef;
		binmode IN;
		# si la taille est dfinie,
		# on ne contrle que les $size premiers octets
		if ($length) {
			read IN, $data, $length;
			close( IN );
	  		$ctx->add( $data );
	  		$sum=$ctx->hexdigest; $length=length( $data );
			}
		# sinon on contrle tout le fichier
		else {
			$ctx->addfile( *IN );
			# to avoid a warning...
			# this variable is meaningless but after an error
			$!=0;
                	close( IN );
                	$sum = $ctx->hexdigest;
			}
	return ($!, $sum);
	}


# purgeArchive ( dbhandle, device, archid )
# remove an archive from the BDD
# in case of a writable disk archive, remove the files too	
sub purgeArchive
	{
	my ($dbhdle, $dev, $purgearch)=@_;
	my (@error, @dberror);
	# THIS SHOULD BE DONE IN A TRANSACTION!!!
	# remove files from the BDD
	my $request="DELETE FROM allfiles WHERE ARCHID=$purgearch";
	my $sth=$dbhdle->do($request) or ( push(@dberror, "[$$] WARNING : Can't delete files from DB! request : $request \n"));
	print "Request : $request \n\n" if ($DEBUG);
	
	# remove the archive from the BDD
	$request="DELETE FROM allarchives WHERE ARCHID=$purgearch";
	$sth=$dbhdle->do($request) or ( push(@dberror, "[$$] WARNING : Can't delete archive $purgearch from DB! request : $request \n"));
	print "Request : $request \n\n" if ($DEBUG);
	
	# remove the archive from disk if necessary
	my $archpath=$dev.'/ARCHID'.$purgearch;
	if ( -e $archpath )
		{
		my $command="rm -rf $archpath";
		print "[$$] issuing \"$command\" \n" if ($VERBOSE);
		my $pid=open3 (\*WRH,\*RDH, \*ERH, $command ) or warn ("[$$] ERROR : couldn't run command : $command !");
		my ($rmerror,$result)=&RunExtCmd($command, "purgeArchive");
		push (@error, $rmerror) if ($rmerror);
		}
	
	return join("\n",@dberror, @error);
	}

# PurgeTape (dbhdle, device, tape)
# erase a tape and all content

sub PurgeTape 
	{
	}

#####################
#        general subs
#####################

#DBquote
sub DBquote
	{
	my ($str)=@_;
	$str=~ s/("|'|\\)/\\$1/g;

	return $str;
	}

# get file or dir size using du
# GeFileSize (file)
sub GetFileSize
	{
	my ($file)=@_;
	$file=~ s/(\W)/\\$1/g;
	my $command="$ConfVar{'DU_prg'} -ks $file";
	my ($error,$result)=&RunExtCmd($command, "GetFileSize");
	if ($result ne '')
		{($result)=split(/\s+/, $result);}
	else
		{$error.="DU got no result for file '$file'.";}
	return ($error, $result);
	}

# get available space on target drive
# GetDriveSpace (dir)
sub GetDriveSpace
	{
	 my ($file)=@_;
	 $file=~ s/(\W)/\\$1/g;
	 my $command="$ConfVar{'DF_prg'} -k $file";
	my ($error,@result)=&RunExtCmd($command, "GetDriveSpace");
	shift(@result);
	my ($fs, $kb)=split(/\s+/, $result[0]);
	return ($error, $kb);
	 }

# send an email report
# SndEmail (destmail, destname, subject, body);
sub SndEmail
	{
	my ($destmail, $destname, $subject, $mailbody)=@_;
	print "dest : $destmail ,$destname\nsubject : $subject\n body : $mailbody\n" if ($VERBOSE);
	print "SMTP host $ConfVar{'smtp_host'} user ".$ENV{USER}." to: $destmail\n" if ($DEBUG);
	my $smtp = Net::SMTP->new($ConfVar{'smtp_host'});
	$smtp->mail($destname);
	$smtp->to($destmail);
	$smtp->data();
	$smtp->datasend("To: $destname\n");
	$smtp->datasend("From: $destname\n");
	$smtp->datasend("Subject: $subject\n");
	$smtp->datasend("\n$mailbody\n");
	$smtp->dataend();
	$smtp->quit;
	}

###############################################
# exits on error, clean up state first...
# ExitOnError (message, job)
sub ExitOnError
	{
	my ($errormessage, $job)=@_;
	my $quotmes=&DBquote($errormessage);
	# set current job to ERROR, set message
	if ($job)
		{
		my $request="UPDATE alljobs set STATUS='ERROR', MESSAGE='$quotmes' where JOBID=$job";
		print "Request : $request \n\n" if ($DEBUG);
		my $sth=$dbhdle->do($request);
		}

	# cleanup tape status :
	# try to write tape end mark if necessary.

	#email the admin
	&SndEmail($ConfVar{'adm_mail'}, '<StorIQ Administrator>', 'StorIQ Archiver encountered an ERROR!', $errormessage);
	warn "ERROR : $errormessage \n";
	$WarningMes='';
	next;
	}

# run external command
sub RunExtCmd
	{
	my ($command, $jobid)=@_;
	print "[$$] issuing $command \n" if ($VERBOSE);
	open (ERH, '>>', "$homedir/../logs/job$jobid.log") or warn ("Can't open job $jobid log for writing: $!");
	my $pid=open3 (\*WRH,\*RDH, "ERH", $command ) or warn ("[$$] ERROR : couldn't run command : $command !");
	my @result=<RDH>;
	close ERH;
	my $error;
	if (!-z "$homedir/../logs/job$jobid.log")
		{
		open (ERH, '<', "$homedir/../logs/job$jobid.log") or warn ("Can't open job $jobid for reading: $!");
		read (ERH, $error, 2048) or warn ("can't read log file job$jobid.log\n");
		close ERH;
		}
	waitpid($pid, 0);
	my $sig=($? & 127);
	$error.="[$$] WARNING : external command '$command' got signal $sig, errno $!\n" if ( $? & 127 );
	$error.="[$$] ERROR : external program CRASHED.\n" if ($? & 128);
	return ($error, @result);
	}
	

