<?php  

class TreeMenu {

/************************************************************

  This treemenu class is based on Bjorge Dijkstra's 
  PHP TreeMenu 1.1. I have wrapped his code up and
  made loading from various sources easier. 

  Bjorge Dijkstra's Original Script:
  -----------------------------------
  http://www.zinc.f2s.com/scriptorium/index.php/treemenu
  
  You can find my adaptation here:
  --------------------------------
  http://www.bitbuddha.com/php/treemenu/treemenu.inc
  http://www.bitbuddha.com/php/treemenu/demo.php

  Example:
  To load the demomenu.txt file supplied with Bjorge's
  distribution, do the following:

  <?php
  $tree = new TreeMenu("a", "demomenu.txt");
  $tree->show();
  ?>

  A global variable $e was used to pass expanded nodes
  from one invocation of the script to the next in Bjorge's
  version. Now you must specify the variable name when
  you create your TreeMenu object:

  $tree = new TreeMenu("a");

  When the object is created, this variable name gets
  the letters "tmv" prepended to it. This allows the
  constructor to loop through the global variables
  and pick out those that it needs when writing certain URLs.

  This came about as a modification to some code sent
  to me by Enrico Stahn [enrico.stahn@gmx.net] on 11/29/2001
  that added multimenu functionality to the script. Much
  thanks to him. I have archived his version at the following URL:

  http://www.bitbuddha.com/php/treemenu/stahn.112901.treemenu.inc  

  I got rid of the overloaded constructor that would accept
  a file name or an array and act accordingly. This required
  programmers to know too much about the implementation
  of the class. I left the ability to pass a file name and supplied
  an addNode() method. You can use it like this:

  <?php
  include "treemenu.inc";
  $tree = new treemenu("a");
  $tree->addNode(1, "grandpa"); 
  $tree->addNode(2, "pa"); 
  $tree->addNode(3, "son"); 
  $tree->addNode(2, "uncle"); 
  $tree->show();
  ?>

  Look at the addNode method signature to see the two optional 
  parameters (ok, so they are $link and $target).

  Your tree will look like this:

  + grandpa
  |
  |----+ pa  
  |    |
  |    |---o son
  |
  |---o uncle

  If you have some hierarchical data in a database
  then it would be pretty easy to write a recursive
  function to build your tree dynamically using the
  addNode() method.

  The loadTreeFromFile() method allows you to add
  as many node files to a tree as you want. You can
  mix it up with other sources (database, etc) using
  the addNode() method, too. An example:

  <?php
  include "treemenu.inc";
  $tree = new TreeMenu("a");
  $tree->addNode(1, "grandpa"); 
  $tree->addNode(2, "pa"); 
  $tree->addNode(3, "son"); 
  $tree->addNode(2, "uncle");
  $tree->loadTreeFromFile("demomenu.txt");
  $tree->show();
  ?>

  There is an additional getHTMLTable() method if
  you ever need that for some reason. It returns
  the table instead of echoing it.

  Denny Shimkoski 2001	dhscompguy@yahoo.com

************************************************************/


	var $tree		= array();
	var $expand		= array();
	var $visible		= array();
	var $levels		= array();
	var $explevels		= array();
	var $urlparams		= array();
	var $maxlevel		= 0;
	var $i			= 0;
	var $urlparam;
	var $script;


	var $img_expand		= "images/tree_expand.gif";
	var $img_collapse	= "images/tree_collapse.gif";
	var $img_line		= "images/tree_vertline.gif";  
	var $img_split		= "images/tree_split.gif";
	var $img_end		= "images/tree_end.gif";
	var $img_leaf		= "images/tree_leaf.gif";
	var $img_spc		= "images/tree_space.gif";

	function treemenu($urlparam, $nodefile = ""){

		$this->script = isset($GLOBALS["PATH_INFO"]) ? $GLOBALS["PATH_INFO"] : $_SERVER["SCRIPT_NAME"];
		$this->urlparam = "tmv".$urlparam;

		reset($GLOBALS["HTTP_GET_VARS"]);
		while (list($key,) = each ($GLOBALS["HTTP_GET_VARS"])){
	      		if (substr($key, 0, 3) == "tmv"){
				$this->urlparams[] = $key;
			}
		}

		if (!empty($nodefile)) $this->loadTreeFromFile($nodefile);
  
	}

	function addNode($level, $text, $link = "", $target = ""){
		$this->tree[$this->i][0] = $level;
		$this->tree[$this->i][1] = "<nobr>".$text."</nobr>";
		$this->tree[$this->i][2] = $link;
		$this->tree[$this->i][3] = $target;
		$this->tree[$this->i][4] = 0;

		if ($this->tree[$this->i][0] > $this->maxlevel) $this->maxlevel = $this->tree[$this->i][0];

		$this->expand[$this->i]=0;
		$this->visible[$this->i]=0;
		$this->levels[$this->i]=0;

		$this->i++;
	}

	function loadTreeFromFile($nodefile){
		$fd = fopen($nodefile, "r");
		if ($fd == 0) die($this->script." : Unable to open file ".$nodefile);
		while ($buffer = fgets($fd, 4096)){
			$level = strspn($buffer,"?");
			$node = explode("|", rtrim(substr($buffer,$level)));
			$text = $node[0];
			$link = $node[1];
			$target = $node[2];
			$this->addNode($level, $text, $link, $target);
		}
		fclose($fd);
	}

	function setExpandedNodesFromURL(){
		if (!empty($GLOBALS[$this->urlparam])) $this->explevels = explode("|",$GLOBALS[$this->urlparam]);
		$this->setExpandedNodes();
	}

	function setExpandedNodes(){
		$i=0;
		while($i<count($this->explevels)){
			$this->expand[$this->explevels[$i]]=1;
			$i++;
		}
	}

	function setEndNodes(){
		$lastlevel=$this->maxlevel;
		for ($i=count($this->tree)-1; $i>=0; $i--){
			if ( $this->tree[$i][0] < $lastlevel ){
				for ($j=$this->tree[$i][0]+1; $j <= $this->maxlevel; $j++){
					$this->levels[$j]=0;
				}
			}

			if ( $this->levels[$this->tree[$i][0]]==0 ){
				$this->levels[$this->tree[$i][0]]=1;
				$this->tree[$i][4]=1;
			} else {
				$this->tree[$i][4]=0;
			}
			$lastlevel=$this->tree[$i][0];
		}
	}

	function setVisibleNodes(){
		// all root nodes are always visible
		for ($i=0; $i < count($this->tree); $i++){
			if ($this->tree[$i][0]==1){
				$this->visible[$i]=1;
			}
		}

		for ($i=0; $i < count($this->explevels); $i++){
			$n = $this->explevels[$i];
			if ( ($this->visible[$n]==1) && ($this->expand[$n]==1) ){
				$j=$n+1;
				while ( $this->tree[$j][0] > $this->tree[$n][0] ){
					if ($this->tree[$j][0]==$this->tree[$n][0]+1) $this->visible[$j]=1;
					$j++;
				}
			}
		}
	}

	function show(){
		echo $this->getHTMLTable();
	}


  	function getHTMLTable(){

		$this->setExpandedNodesFromURL();
		$this->setEndNodes();
		$this->setVisibleNodes();

		/*********************************************
		 *  Output nicely formatted tree             *
		 *********************************************/

		for ($i=0; $i<$this->maxlevel; $i++) $this->levels[$i]=1;

		$this->maxlevel++;
  
		$html = "<table cellspacing=0 cellpadding=0 border=0 cols=".($this->maxlevel+3)." width=100%>\n";
		$html .= "<tr>";
		for ($i=0; $i<$this->maxlevel; $i++) $html .= "<td width=16></td>";
		$html .= "<td width=100%>&nbsp;</td></tr>\n";

		$cnt=0;

		while ($cnt<count($this->tree)){
			if ($this->visible[$cnt]){

		/****************************************
		 * start new row                        *
		 ****************************************/

				$html .= "<tr>";
      
		/****************************************
		 * vertical lines from higher levels    *
		 ****************************************/

				$i=0;
				while ($i<$this->tree[$cnt][0]-1){
					if ($this->levels[$i]==1){
						$html .= "<td><a name='$cnt'></a><img src=\"".$this->img_line."\"></td>";
					} else {
						$html .= "<td><a name='$cnt'></a><img src=\"".$this->img_spc."\"></td>";
					}
					$i++;
				}
      
		/****************************************
		 * corner at end of subtree or t-split  *
		 ****************************************/         

				if ($this->tree[$cnt][4]==1){
					$html .= "<td><img src=\"".$this->img_end."\"></td>";
					$this->levels[$this->tree[$cnt][0]-1]=0;
				} else {
					$html .= "<td><img src=\"".$this->img_split."\"></td>";                  
					$this->levels[$this->tree[$cnt][0]-1]=1;    
				}
      
		/********************************************
		 * Node (with subtree) or Leaf (no subtree) *
		 ********************************************/

				if ($this->tree[$cnt+1][0]>$this->tree[$cnt][0]){
        
		/****************************************
		 * Create expand/collapse parameters    *
		 ****************************************/

					reset($this->urlparams);
					$otherparams = "";
					while (list(,$value) = each ($this->urlparams)){
	     					if (!empty($GLOBALS[$value]) && $value <> $this->urlparam) {
	     						$otherparams .= $value."=".$GLOBALS[$value]."&";
	     					}
					}
					
					$params="?".$otherparams.$this->urlparam."=";

					$i=0;
					while($i<count($this->expand)){
						if ( ($this->expand[$i]==1) 
							&& ($cnt!=$i) 
							|| ($this->expand[$i]==0 && $cnt==$i)){

							$params=$params.$i;
							$params=$params."|";
						}
						$i++;
					}
               
					$otherparams = "";

					if ($this->expand[$cnt]==0){
						$html .= "<td><a href=\"".$this->script.$params."#$cnt\"><img src=\"".$this->img_expand."\" border=no></a></td>";
					} else {
						$html .= "<td><a href=\"".$this->script.$params."#$cnt\"><img src=\"".$this->img_collapse."\" border=no></a></td>";
					}

				} else {

		/*************************
		 * Tree Leaf             *
		 *************************/

					$html .= "<td><img src=\"".$this->img_leaf."\"></td>";         
				}
      
		/****************************************
		 * output item text                     *
		 ****************************************/

				if ($this->tree[$cnt][2]==""){
					$html .= "<td colspan=".($this->maxlevel-$this->tree[$cnt][0]).">".$this->tree[$cnt][1]."</td>";
				} else {
					$html .= "<td colspan=".($this->maxlevel-$this->tree[$cnt][0])."><a href=\"".$this->tree[$cnt][2]."\" target=\"".$this->tree[$cnt][3]."\">".$this->tree[$cnt][1]."</a></td>";
				}

		/****************************************
		 * end row                              *
		 ****************************************/
              
				$html .= "</tr>\n";
			}

			$cnt++;
		}

		$html .= "</table>\n";
		return $html;
	}
}


?>

